@@ -4,15 +4,10 @@
 # later version. See the file COPYING for details.
 
 
-try:
-    #new recommended way of using the glib main loop:
-    from dbus.mainloop.glib import DBusGMainLoop
-    DBusGMainLoop(set_as_default=True)
-except:
-    #beware: this import has side-effects:
-    import dbus.glib
-    assert dbus.glib
-import dbus #@Reimport
+#new recommended way of using the glib main loop:
+from dbus.mainloop.glib import DBusGMainLoop
+DBusGMainLoop(set_as_default=True)
+import dbus  #pylint: disable=wrong-import-position
 
 
 #NAME = "org.freedesktop.ScreenSaver"
@@ -27,9 +22,8 @@ def main():
         def active_changed(active):
             print("screensaver active status changed: %s" % (active, ))
         dbus_session.add_signal_receiver(active_changed, "ActiveChanged", NAME, path=PATH)
-        from xpra.gtk_common.gobject_compat import import_glib
-        glib = import_glib()
-        loop = glib.MainLoop()
+        from gi.repository import GLib  # @UnresolvedImport
+        loop = GLib.MainLoop()
         loop.run()
 
 
