@@ -1,4 +1,4 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 
 import sys
 
@@ -16,7 +16,9 @@ def notify(*_args):
     hints = {
         "image-path"    : "/usr/share/xpra/icons/encoding.png",
         }
-    notifier.show_notify("dbus-id", None, nid, "xpra test app", 0, "", "Notification %i Summary" % nid, "Notification %i Body" % nid, actions, hints, 60*1000, "")
+    notifier.show_notify("dbus-id", None, nid, "xpra test app", 0, "",
+                         "Notification %i Summary" % nid, "Notification %i Body" % nid,
+                         actions, hints, 60*1000, "")
     nid += 1
     return True
 
@@ -28,10 +30,9 @@ def main():
     from xpra.platform import program_context
     with program_context(name, name):
         notify()
-        from xpra.gtk_common.gobject_compat import import_glib
-        glib = import_glib()
-        glib.timeout_add(60*1000, notify)
-        loop = glib.MainLoop()
+        from gi.repository import GLib  # @UnresolvedImport
+        GLib.timeout_add(60*1000, notify)
+        loop = GLib.MainLoop()
         loop.run()
 
 
