@@ -1,6 +1,6 @@
 # This file is part of Xpra.
 # Copyright (C) 2008, 2009 Nathaniel Smith <njs@pobox.com>
-# Copyright (C) 2012-2019 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2012-2023 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -24,38 +24,39 @@
 # do another XSync).
 #
 # (In this modern world, with WM's either on the same machine or over
-# super-fast connections to the X server, everything running on fast
+# superfast connections to the X server, everything running on fast
 # computers... does being this careful to avoid sync's actually matter?)
 
-import threading
 import traceback
+from typing import Dict
+
+import gi
+gi.require_version("Gdk", "3.0")  # @UndefinedVariable
+from gi.repository import Gdk  # @UnresolvedImport
 
 from xpra.util import envbool
-from xpra.os_util import bytestostr
-from xpra.gtk_common.gobject_compat import import_gdk
+from xpra.os_util import is_main_thread
 from xpra.log import Logger
 
-__all__ = ["XError", "trap", "xsync", "xswallow"]
+__all__ = ["XError", "trap", "xsync", "xswallow", "xlog", "verify_sync"]
 
 #run xpra in synchronized mode to debug X11 errors:
 XPRA_SYNCHRONIZE = envbool("XPRA_SYNCHRONIZE", False)
 XPRA_LOG_SYNC = envbool("XPRA_LOG_SYNC", False)
 VERIFY_MAIN_THREAD = envbool("XPRA_VERIFY_MAIN_THREAD", True)
+LOG_NESTED_XTRAP = envbool("XPRA_LOG_NESTED_XTRAP", False)
 
 log = Logger("x11", "util")
 elog = Logger("x11", "util", "error")
 
-gdk = import_gdk()
-
 
 if not VERIFY_MAIN_THREAD:
     def verify_main_thread():
         return
 else:
-    main_thread = threading.current_thread()
     def verify_main_thread():
-        ct = threading.current_thread()
-        if main_thread != ct:
+        if not is_main_thread():
+            import threading
             log.error("Error: invalid access from thread %s", threading.current_thread())
             traceback.print_stack()
     verify_main_thread()
@@ -63,53 +64,50 @@ else:
 
 class XError(Exception):
     def __init__(self, message):
-        Exception.__init__(self)
-        self.msg = message
+        super().__init__()
+        self.msg = get_X_error(message)
 
     def __str__(self):
-        return "XError: %s" % bytestostr(self.msg)
+        return "XError: %s" % self.msg
 
 
-xerror_to_name = None
-def get_X_error(xerror):
+xerror_to_name : Dict[int,str] = {}
+def get_X_error(xerror) -> str:
     global xerror_to_name
     if not isinstance(xerror, int):
-        return xerror
+        return str(xerror)
     try:
-        from xpra.x11.bindings.window_bindings import constants     #@UnresolvedImport
-        if xerror_to_name is None:
-            xerror_to_name = {}
-            for name,code in constants.items():
+        from xpra.x11.bindings.window import constants     #@UnresolvedImport
+        if not xerror_to_name:
+            xerror_to_name[0] = "OK"
+            for name,code in constants.items():  # @UndefinedVariable
                 if name=="Success" or name.startswith("Bad"):
                     xerror_to_name[code] = name
             log("get_X_error(..) initialized error names: %s", xerror_to_name)
         if xerror in xerror_to_name:
-            return xerror_to_name.get(xerror)
-        from xpra.x11.bindings.core_bindings import X11CoreBindings     #@UnresolvedImport
+            return xerror_to_name.get(xerror) or str(xerror)
+        from xpra.x11.bindings.core import X11CoreBindings     #@UnresolvedImport
         return X11CoreBindings().get_error_text(xerror)
     except Exception as e:
         log.error("get_X_error(%s) %s", xerror, e, exc_info=True)
-    return xerror
-
-class XErrorInfo(object):
-    def __init__(self, xerror):
-        self.xerror = xerror
-    def __repr__(self):
-        return str(get_X_error(self.xerror))
+    return str(xerror)
 
 
 # gdk has its own depth tracking stuff, but we have to duplicate it here to
 # minimize calls to XSync.
-class _ErrorManager(object):
+class _ErrorManager:
     def __init__(self):
         self.depth = 0
 
     def Xenter(self):
         assert self.depth >= 0
         verify_main_thread()
-        gdk.error_trap_push()
+        Gdk.error_trap_push()
         if XPRA_LOG_SYNC:
             log("X11trap.enter at level %i", self.depth)
+        if LOG_NESTED_XTRAP and self.depth>0:
+            for x in traceback.extract_stack():
+                log("%s", x)
         self.depth += 1
 
     def Xexit(self, need_sync=True):
@@ -118,11 +116,17 @@ class _ErrorManager(object):
         if XPRA_LOG_SYNC:
             log("X11trap.exit at level %i, need_sync=%s", self.depth, need_sync)
         if self.depth == 0 and need_sync:
-            gdk.flush()
+            Gdk.flush()
         # This is a Xlib error constant (Success == 0)
-        error = gdk.error_trap_pop()
+        error = Gdk.error_trap_pop()
         if error:
-            raise XError(get_X_error(error))
+            raise XError(error)
+
+    def safe_x_exit(self):
+        try:
+            self.Xexit()
+        except XError as e:
+            log(f"Warning: '{e}' detected while already in unwind; discarding")
 
     def _call(self, need_sync, fun, args, kwargs):
         # Goal: call the function.  In all conditions, call _exit exactly once
@@ -139,7 +143,7 @@ class _ErrorManager(object):
             try:
                 self.Xexit(need_sync)
             except XError as ee:
-                log("XError %s detected while already in unwind; discarding", XErrorInfo(ee))
+                log(f"XError '{ee}' detected while already in unwind; discarding")
             raise
         self.Xexit(need_sync)
         return value
@@ -159,16 +163,18 @@ class _ErrorManager(object):
         try:
             self.call_unsynced(fun, *args, **kwargs)
             return True
-        except XError as e:
-            log("Ignoring X error: %s on %s", XErrorInfo(e.msg), fun)
+        except XError:
+            log("Ignoring X error on %s",
+                fun, exc_info=True)
             return False
 
     def swallow_synced(self, fun, *args, **kwargs):
         try:
             self.call_synced(fun, *args, **kwargs)
             return True
-        except XError as e:
-            log("Ignoring X error: %s on %s", XErrorInfo(e.msg), fun)
+        except XError:
+            log("Ignoring X error on %s",
+                fun, exc_info=True)
             return False
 
     if XPRA_SYNCHRONIZE:
@@ -182,65 +188,60 @@ class _ErrorManager(object):
 trap = _ErrorManager()
 
 
-class XSyncContext(object):
+class XSyncContext:
 
     def __enter__(self):
         trap.Xenter()
 
-    def __exit__(self, e_typ, _e_val, _trcbak):
-        #log("xsync.exit%s", (e_typ, e_val, trcbak))
+    def __exit__(self, e_typ, _e_val, trcbak):
         try:
             trap.Xexit()
-        except XError as ee:
+        except XError as e:
             if e_typ is None:
                 #we are not handling an exception yet, so raise this one:
                 raise
-            log("XError %s detected while already in unwind; discarding", XErrorInfo(ee))
+            log(f"Ignoring {e} during Xexit, {e_typ} will be raised instead", exc_info=e)
         #raise the original exception:
         return False
 
 xsync = XSyncContext()
 
 
-class XSwallowContext(object):
+class XSwallowContext:
 
     def __enter__(self):
         trap.Xenter()
 
     def __exit__(self, e_typ, e_val, trcbak):
         if e_typ:
-            log("xswallow.exit%s", (e_typ, e_val, trcbak), exc_info=True)
-        try:
-            trap.Xexit()
-        except XError as ee:
-            log("XError %s detected while already in unwind; discarding", XErrorInfo(ee))
+            log("XError swallowed: %s, %s", e_typ, e_val, exc_info=trcbak)
+        trap.safe_x_exit()
         #don't raise exceptions:
         return True
 
 xswallow = XSwallowContext()
 
 
-class XLogContext(object):
+class XLogContext:
 
     def __enter__(self):
         trap.Xenter()
 
     def __exit__(self, e_typ, e_val, trcbak):
         if e_typ:
-            log.error("XError: %s, %s", XErrorInfo(e_typ), e_val, exc_info=True)
-        try:
-            trap.Xexit()
-        except XError as ee:
-            log("XError %s detected while already in unwind; discarding", XErrorInfo(ee))
+            log.error("Error: %s, %s", e_typ, e_val, exc_info=trcbak)
+        trap.safe_x_exit()
         #don't raise exceptions:
         return True
 
 xlog = XLogContext()
 
 
-def verify_sync():
+def verify_sync(*args):
     if trap.depth<=0:
         log.error("Error: unmanaged X11 context")
+        if args:
+            log.error(" " + str(args[0]) + "%s", args[1:])
         stack = traceback.extract_stack()[:-1]
         s = traceback.format_list(stack)
         for x in s:
