@@ -1,70 +1,48 @@
 # This file is part of Xpra.
-# Copyright (C) 2011-2019 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2011-2023 Antoine Martin <antoine@xpra.org>
 # Copyright (C) 2008, 2009, 2010 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 import sys
 import os
-import time
 import errno
 import socket
+from typing import Dict, Any, Optional, Union, Callable
 
-from xpra.net.common import ConnectionClosedException
-from xpra.util import envint, envbool, csv
-from xpra.os_util import WIN32, PYTHON2, POSIX, LINUX
+from xpra.net.common import ConnectionClosedException, IP_SOCKTYPES, TCP_SOCKTYPES
+from xpra.util import envint, envbool, hasenv, csv
+from xpra.common import FULL_INFO
+from xpra.make_thread import start_thread
+from xpra.os_util import POSIX, LINUX, WIN32, OSX
 from xpra.platform.features import TCP_OPTIONS, IP_OPTIONS, SOCKET_OPTIONS
 from xpra.log import Logger
 
 log = Logger("network", "protocol")
 
-SOCKET_CORK = envbool("XPRA_SOCKET_CORK", LINUX)
+SOCKET_CORK : bool = envbool("XPRA_SOCKET_CORK", LINUX)
 if SOCKET_CORK:
     try:
-        assert socket.TCP_CORK>0
+        assert socket.TCP_CORK>0  #@UndefinedVariable
     except (AttributeError, AssertionError) as cork_e:
         log.warn("Warning: unable to use TCP_CORK on %s", sys.platform)
         log.warn(" %s", cork_e)
         SOCKET_CORK = False
-SOCKET_NODELAY = envbool("XPRA_SOCKET_NODELAY", None)
-VSOCK_TIMEOUT = envint("XPRA_VSOCK_TIMEOUT", 5)
-SOCKET_TIMEOUT = envint("XPRA_SOCKET_TIMEOUT", 20)
-SSL_PEEK = envbool("XPRA_SSL_PEEK", True)
+SOCKET_NODELAY : Optional[bool] = None
+if hasenv("XPRA_SOCKET_NODELAY"):
+    SOCKET_NODELAY = envbool("XPRA_SOCKET_NODELAY")
+SOCKET_KEEPALIVE : bool = envbool("XPRA_SOCKET_KEEPALIVE", True)
+VSOCK_TIMEOUT : int = envint("XPRA_VSOCK_TIMEOUT", 5)
+SOCKET_TIMEOUT : int = envint("XPRA_SOCKET_TIMEOUT", 20)
 #this is more proper but would break the proxy server:
-SOCKET_SHUTDOWN = envbool("XPRA_SOCKET_SHUTDOWN", False)
-LOG_TIMEOUTS = envint("XPRA_LOG_TIMEOUTS", 1)
+SOCKET_SHUTDOWN : bool = envbool("XPRA_SOCKET_SHUTDOWN", False)
+LOG_TIMEOUTS : int = envint("XPRA_LOG_TIMEOUTS", 1)
 
-#on some platforms (ie: OpenBSD), reading and writing from sockets
-#raises an IOError but we should continue if the error code is EINTR
-#this wrapper takes care of it.
-#EWOULDBLOCK can also be hit with the proxy server when we handover the socket
-CONTINUE_ERRNO = {
-            errno.EINTR         : "EINTR",
-            errno.EWOULDBLOCK   : "EWOULDBLOCK"
-            }
-ABORT = {
+ABORT : Dict[int, str] = {
          errno.ENXIO            : "ENXIO",
          errno.ECONNRESET       : "ECONNRESET",
          errno.EPIPE            : "EPIPE",
          }
-continue_wait = 0
-
-#default to using os.read and os.write for both tty devices and regular streams
-#(but overriden for win32 below for tty devices to workaround an OS "feature")
-OS_READ = os.read
-OS_WRITE = os.write
-TTY_READ = os.read
-TTY_WRITE = os.write
-if WIN32 and PYTHON2:
-    #win32 has problems writing more than 32767 characters to stdout!
-    #see: http://bugs.python.org/issue11395
-    #(this is fixed in python 3.2 and we don't care about 3.0 or 3.1)
-    def win32ttywrite(fd, buf):
-        #this awful limitation only applies to tty devices:
-        if len(buf)>32767:
-            buf = buf[:32767]
-        return os.write(fd, buf)
-    TTY_WRITE = win32ttywrite
 
 
 PROTOCOL_STR = {}
@@ -74,73 +52,52 @@ for x in dir(socket):
         PROTOCOL_STR[getattr(socket, x)] = x
     if x.startswith("SOCK_"):
         FAMILY_STR[getattr(socket, x)] = x
-
-
-def set_continue_wait(v):
-    global continue_wait
-    continue_wait = v
+del x
 
 CAN_RETRY_EXCEPTIONS = ()
 CLOSED_EXCEPTIONS = ()
 
-if PYTHON2:
-    from io import BlockingIOError
-else:
-    assert BlockingIOError
-
-def can_retry(e):
+def can_retry(e) -> Union[bool,str]:
     if isinstance(e, socket.timeout):
         return "socket.timeout"
     if isinstance(e, BlockingIOError):
         return True
-    if isinstance(e, (IOError, OSError)):
-        global CONTINUE_ERRNO
-        code = e.args[0]
-        can_continue = CONTINUE_ERRNO.get(code)
-        if can_continue:
-            return can_continue
-
-        #SSL pollution - see ticket #1927
-        if code=="The read operation timed out":
-            return str(code)
-
+    if isinstance(e, BrokenPipeError):
+        raise ConnectionClosedException(e) from None
+    if isinstance(e, OSError):
         if isinstance(e, CAN_RETRY_EXCEPTIONS):
             return str(e)
-
+        code = e.args[0]
         abort = ABORT.get(code, code)
         if abort is not None:
             err = getattr(e, "errno", None)
             log("can_retry: %s, args=%s, errno=%s, code=%s, abort=%s", type(e), e.args, err, code, abort)
-            raise ConnectionClosedException(e)
+            raise ConnectionClosedException(e) from None
     if isinstance(e, CLOSED_EXCEPTIONS):
-        raise ConnectionClosedException(e)
+        raise ConnectionClosedException(e) from None
     return False
 
-def untilConcludes(is_active_cb, can_retry_cb, f, *a, **kw):
-    global continue_wait
-    wait = 0
+def untilConcludes(is_active_cb:Callable, can_retry_cb:Callable, f:Callable, *a, **kw):
     while is_active_cb():
         try:
             return f(*a, **kw)
         except Exception as e:
             retry = can_retry_cb(e)
+            if not retry:
+                raise
             if LOG_TIMEOUTS>0:
                 log("untilConcludes(%s, %s, %s, %s, %s) %s, retry=%s",
                     is_active_cb, can_retry_cb, f, a, kw, e, retry, exc_info=LOG_TIMEOUTS>=2)
-            e = None
-            if not retry:
-                raise
-            if wait>0:
-                time.sleep(wait/1000.0)     #wait is in milliseconds, sleep takes seconds
-            if wait<continue_wait:
-                wait += 1
 
 
-def pretty_socket(s):
+def pretty_socket(s) -> str:
     try:
         if isinstance(s, str):
             return s
         if len(s)==2:
+            if str(s[0]).find(":")>=0:
+                #IPv6
+                return "[%s]:%s" % (s[0], s[1])
             return "%s:%s" % (s[0], s[1])
         if len(s)==4:
             return csv(str(x) for x in s)
@@ -149,9 +106,9 @@ def pretty_socket(s):
     return str(s)
 
 
-class Connection(object):
-    def __init__(self, endpoint, socktype, info=None):
-        log("Connection%s", (endpoint, socktype, info))
+class Connection:
+    def __init__(self, endpoint, socktype, info=None, options=None):
+        log("Connection%s", (endpoint, socktype, info, options))
         self.endpoint = endpoint
         try:
             assert isinstance(endpoint, (tuple, list))
@@ -161,6 +118,7 @@ class Connection(object):
         self.socktype_wrapped = socktype
         self.socktype = socktype
         self.info = info or {}
+        self.options = options or {}
         self.input_bytecount = 0
         self.input_readcount = 0
         self.output_bytecount = 0
@@ -169,32 +127,32 @@ class Connection(object):
         self.active = True
         self.timeout = 0
 
-    def set_nodelay(self, nodelay):
-        pass
+    def set_nodelay(self, nodelay : bool) -> None:
+        """ TCP sockets override this method  """
 
-    def set_cork(self, cork):
-        pass
+    def set_cork(self, cork : bool) -> None:
+        """ TCP sockets override this method  """
 
-    def is_active(self):
+    def is_active(self) -> bool:
         return self.active
 
-    def set_active(self, active):
+    def set_active(self, active : bool) -> None:
         self.active = active
 
-    def close(self):
+    def close(self) -> None:
         self.set_active(False)
 
-    def can_retry(self, e):
+    def can_retry(self, e) -> Union[bool,str]:
         return can_retry(e)
 
     def untilConcludes(self, *args):
         return untilConcludes(self.is_active, self.can_retry, *args)
 
-    def peek(self, _n):
+    def peek(self, _n : int) -> bytes:
         #not implemented
-        return None
+        return b""
 
-    def _write(self, *args):
+    def _write(self, *args) -> int:
         """ wraps do_write with packet accounting """
         w = self.untilConcludes(*args)
         self.output_bytecount += w or 0
@@ -208,7 +166,7 @@ class Connection(object):
         self.input_readcount += 1
         return r
 
-    def get_info(self):
+    def get_info(self) -> Dict[str,Any]:
         info = self.info.copy()
         if self.socktype_wrapped!=self.socktype:
             info["wrapped"] = self.socktype_wrapped
@@ -234,19 +192,11 @@ class Connection(object):
 # socket (if it exists)
 class TwoFileConnection(Connection):
     def __init__(self, writeable, readable, abort_test=None, target=None, socktype="", close_cb=None, info=None):
-        Connection.__init__(self, target, socktype, info)
+        super().__init__(target, socktype, info)
         self._writeable = writeable
         self._readable = readable
         self._read_fd = self._readable.fileno()
         self._write_fd = self._writeable.fileno()
-        if os.isatty(self._read_fd):
-            self._osread = TTY_READ
-        else:
-            self._osread = OS_READ
-        if os.isatty(self._write_fd):
-            self._oswrite = TTY_WRITE
-        else:
-            self._oswrite = OS_WRITE
         self._abort_test = abort_test
         self._close_cb = close_cb
 
@@ -255,38 +205,55 @@ class TwoFileConnection(Connection):
         if self._abort_test:
             self._abort_test(action)
 
+    def flush(self) -> None:
+        r = self._readable
+        if r:
+            r.flush()
+        w = self._writeable
+        if w:
+            w.flush()
+
     def read(self, n):
         self.may_abort("read")
-        return self._read(self._osread, self._read_fd, n)
+        return self._read(os.read, self._read_fd, n)
 
-    def write(self, buf):
+    def write(self, buf, _packet_type:str=""):
         self.may_abort("write")
-        return self._write(self._oswrite, self._write_fd, buf)
+        return self._write(os.write, self._write_fd, buf)
 
-    def close(self):
+    def close(self) -> None:
         log("%s.close() close callback=%s, readable=%s, writeable=%s",
             self, self._close_cb, self._readable, self._writeable)
-        Connection.close(self)
+        super().close()
         cc = self._close_cb
         if cc:
             self._close_cb = None
             log("%s.close() calling %s", self, cc)
-            cc()
-        try:
-            self._readable.close()
-        except IOError as e:
-            log("%s.close() %s", self._readable, e)
-        try:
-            self._writeable.close()
-        except IOError as e:
-            log("%s.close() %s", self._writeable, e)
+            try:
+                cc()
+            except Exception:
+                log.error("%s.close() error on callback %s", self, cc, exc_info=True)
+        def close_files_thread() -> None:
+            log("close_files_thread() _readable=%s", self._readable)
+            log("close_files_thread() calling %s", self._readable.close)
+            try:
+                self._readable.close()
+            except IOError as e:
+                log("close_files_thread() %s", self._readable, e)
+            log("close_files_thread() _writeable=%s", self._writeable)
+            log("close_files_thread() calling %s", self._writeable.close)
+            try:
+                self._writeable.close()
+            except IOError as e:
+                log("close_files_thread() %s", self._writeable, e)
+        start_thread(close_files_thread, "close-files-thread", daemon=True)
         log("%s.close() done", self)
 
     def __repr__(self):
-        return "Pipe(%s)" % str(self.target)
+        return f"Pipe({self.target})"
 
-    def get_info(self):
-        d = Connection.get_info(self)
+    def get_info(self) -> Dict[str,Any]:
+        d = super().get_info()
         d.update({
             "type"  : "pipe",
             "pipe"  : {
@@ -297,56 +264,100 @@ class TwoFileConnection(Connection):
         return d
 
 
-TCP_SOCKTYPES = ("tcp", "ssl", "ws", "wss")
 
 class SocketConnection(Connection):
-    def __init__(self, sock, local, remote, target, socktype, info=None):
-        log("SocketConnection%s", (sock, local, remote, target, socktype, info))
-        Connection.__init__(self, target, socktype, info)
+    def __init__(self, sock, local, remote, target, socktype, info=None, socket_options=None):
+        log("SocketConnection%s", (sock, local, remote, target, socktype, info, socket_options))
+        super().__init__(target, socktype, info, socket_options)
         self._socket = sock
         self.local = local
         self.remote = remote
         self.protocol_type = "socket"
-        self.nodelay = None
-        self.cork = None
+        if self.socktype_wrapped in TCP_SOCKTYPES:
+            def boolget(k, default_value):
+                v = self.options.get(k)
+                if v is None:
+                    return default_value
+                try:
+                    return bool(int(v))
+                except ValueError:
+                    return default_value
+            self.cork = boolget("cork", SOCKET_CORK)
+            self.nodelay = boolget("nodelay", SOCKET_NODELAY)
+            log("%s options: cork=%s, nodelay=%s", self.socktype_wrapped, self.cork, self.nodelay)
+            if self.nodelay:
+                self.do_set_nodelay(self.nodelay)
+            keepalive = boolget("keepalive", SOCKET_KEEPALIVE)
+            try:
+                self._setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, int(keepalive))
+                if keepalive:
+                    idletime = 10
+                    interval = 3
+                    kmax = 5
+                    if WIN32:
+                        sock = self.get_raw_socket()
+                        if sock:
+                            #@UndefinedVariable pylint: disable=no-member
+                            sock.ioctl(socket.SIO_KEEPALIVE_VALS, (1, idletime*1000, interval*1000))
+                    elif OSX:
+                        TCP_KEEPALIVE = 0x10
+                        self._setsockopt(socket.IPPROTO_TCP, TCP_KEEPALIVE, interval)
+                    elif LINUX:
+                        self._setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, idletime)
+                        self._setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, interval)
+                        self._setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, kmax)
+            except OSError:
+                log("cannot set KEEPALIVE", exc_info=True)
+        else:
+            self.cork = False
+            self.nodelay = False
+        self.nodelay_value = None
+        self.cork_value = None
         if isinstance(remote, str):
             self.filename = remote
-        if SOCKET_NODELAY is not None and self.socktype in TCP_SOCKTYPES:
-            self.do_set_nodelay(SOCKET_NODELAY)
 
-    def set_nodelay(self, nodelay):
-        if SOCKET_NODELAY is None and self.socktype_wrapped in TCP_SOCKTYPES and self.nodelay!=nodelay:
+    def get_raw_socket(self):
+        return self._socket
+
+    def _setsockopt(self, *args) -> None:
+        if self.active:
+            sock = self.get_raw_socket()
+            if sock:
+                sock.setsockopt(*args)
+
+    def set_nodelay(self, nodelay : bool) -> None:
+        if self.nodelay is None and self.nodelay_value!=nodelay:
             self.do_set_nodelay(nodelay)
 
-    def do_set_nodelay(self, nodelay):
-        self._socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, nodelay)
-        self.nodelay = nodelay
+    def do_set_nodelay(self, nodelay : bool) -> None:
+        self._setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, nodelay)
+        self.nodelay_value = nodelay
         log("changed %s socket to nodelay=%s", self.socktype, nodelay)
 
-    def set_cork(self, cork):
-        if SOCKET_CORK and self.socktype_wrapped in TCP_SOCKTYPES and self.cork!=cork:
-            self._socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_CORK, cork)
-            self.cork = cork
+    def set_cork(self, cork : bool) -> None:
+        if self.cork and self.cork_value!=cork:
+            self._setsockopt(socket.IPPROTO_TCP, socket.TCP_CORK, cork)  #@UndefinedVariable
+            self.cork_value = cork
             log("changed %s socket to cork=%s", self.socktype, cork)
 
-    def peek(self, n):
+    def enable_peek(self, peeked=b""):
+        assert not isinstance(self._socket, SocketPeekWrapper)
+        self._socket = SocketPeekWrapper(self._socket, peeked)
+
+    def peek(self, n : int) -> bytes:
+        log("%s(%s, MSG_PEEK)", self._socket.recv, n)
         return self._socket.recv(n, socket.MSG_PEEK)
 
-    def read(self, n):
+    def read(self, n : int) -> bytes:
         return self._read(self._socket.recv, n)
 
-    def write(self, buf):
+    def write(self, buf, _packet_type:str=""):
         return self._write(self._socket.send, buf)
 
-    def close(self):
+    def close(self) -> None:
         s = self._socket
-        try:
-            i = self.get_socket_info()
-        except IOError:
-            i = s
-        log("%s.close() for socket=%s", self, i)
-        Connection.close(self)
-        #meaningless for udp:
+        log(f"{self}.close() socket={s}")
+        super().close()
         try:
             s.settimeout(0)
         except IOError:
@@ -367,7 +378,7 @@ class SocketConnection(Connection):
                 raise
         log("%s.close() done", self)
 
-    def error_is_closed(self, e):
+    def error_is_closed(self, e) -> bool:
         return isinstance(e, CLOSED_EXCEPTIONS)
 
     def __repr__(self):
@@ -376,32 +387,34 @@ class SocketConnection(Connection):
                 self.socktype, self.protocol_type,
                 pretty_socket(self.local), pretty_socket(self.remote),
                 )
-        return "%s %s:%s" % (self.socktype, self.protocol_type, pretty_socket(self.local))
+        return f"{self.socktype} {self.protocol_type}:{pretty_socket(self.local)}"
 
-    def get_info(self):
-        d = Connection.get_info(self)
+    def get_info(self) -> Dict[str,Any]:
+        d = super().get_info()
         try:
             d["remote"] = self.remote or ""
             d["protocol-type"] = self.protocol_type
-            si = self.get_socket_info()
-            if si:
-                d["socket"] = si
+            if FULL_INFO>0:
+                si = self.get_socket_info()
+                if si:
+                    d["socket"] = si
         except socket.error:
             log.error("Error accessing socket information", exc_info=True)
         return d
 
-    def get_socket_info(self):
+    def get_socket_info(self) -> Dict[str,Any]:
         return self.do_get_socket_info(self._socket)
 
-    def do_get_socket_info(self, s):
+    def do_get_socket_info(self, s) -> Dict[str,Any]:
         if not s:
-            return None
+            return {}
         info = {}
         try:
             info.update({
                 "proto"         : s.proto,
                 "family"        : FAMILY_STR.get(s.family, int(s.family)),
                 "type"          : PROTOCOL_STR.get(s.type, int(s.type)),
+                "cork"          : self.cork,
                 })
         except AttributeError:
             log("do_get_socket_info()", exc_info=True)
@@ -418,28 +431,44 @@ class SocketConnection(Connection):
                 fd = 0
             if fd:
                 info["fileno"] = fd
-            from xpra.platform.netdev_query import get_interface_info
             #ie: self.local = ("192.168.1.7", "14500")
-            if self.local and len(self.local)==2:
+            log("do_get_socket_info(%s) fd=%s, local=%s", s, fd, self.local)
+            if self.local and len(self.local)==2 and FULL_INFO > 1:
                 from xpra.net.net_util import get_interface
                 iface = get_interface(self.local[0])
                 #ie: iface = "eth0"
+                device_info = {}
+                if iface:
+                    device_info["name"] = iface
                 if iface and iface!="lo":
-                    i = get_interface_info(fd, iface)
-                    if i:
-                        info["device"] = i
-        except OSError as e:
+                    try:
+                        from xpra.platform.netdev_query import get_interface_info
+                    except ImportError as e:
+                        log("do_get_socket_info() no netdev_query: %s", e)
+                    else:
+                        device_info.update(get_interface_info(fd, iface))
+                if device_info:
+                    info["device"] = device_info
+        except (OSError, ValueError) as e:
             log("do_get_socket_info() error querying socket speed", exc_info=True)
             log.error("Error querying socket speed:")
-            log.error(" %s", e)
+            log.estr(e)
         else:
             opts = {
                     "SOCKET" : get_socket_options(s, socket.SOL_SOCKET, SOCKET_OPTIONS),
                     }
-            if self.socktype_wrapped in ("tcp", "udp", "ws", "wss", "ssl"):
+            if self.socktype_wrapped in IP_SOCKTYPES:
                 opts["IP"] = get_socket_options(s, socket.SOL_IP, IP_OPTIONS)
-            if self.socktype_wrapped in ("tcp", "ws", "wss", "ssl"):
+            if self.socktype_wrapped in TCP_SOCKTYPES:
                 opts["TCP"] = get_socket_options(s, socket.IPPROTO_TCP, TCP_OPTIONS)
+                from xpra.platform.netdev_query import get_tcp_info
+                try:
+                    opts["TCP_INFO"] = get_tcp_info(s)
+                except (OSError, ValueError) as e:
+                    log(f"get_tcp_info({s})", exc_info=True)
+                    if self.is_active() and not self.error_is_closed(e):
+                        log.warn("Warning: failed to get tcp information")
+                        log.warn(f" from {self.socktype} socket {self}")
             #ipv6:  IPV6_ADDR_PREFERENCES, IPV6_CHECKSUM, IPV6_DONTFRAG, IPV6_DSTOPTS, IPV6_HOPOPTS,
             # IPV6_MULTICAST_HOPS, IPV6_MULTICAST_IF, IPV6_MULTICAST_LOOP, IPV6_NEXTHOP, IPV6_PATHMTU,
             # IPV6_PKTINFO, IPV6_PREFER_TEMPADDR, IPV6_RECVDSTOPTS, IPV6_RECVHOPLIMIT, IPV6_RECVHOPOPTS,
@@ -449,7 +478,7 @@ class SocketConnection(Connection):
         return info
 
 
-def get_socket_options(sock, level, options):
+def get_socket_options(sock, level, options) -> Dict:
     opts = {}
     errs = []
     for k in options:
@@ -465,14 +494,19 @@ def get_socket_options(sock, level, options):
             if v is not None:
                 opts[k] = v
     if errs:
-        log.warn("Warning: failed to query %s", csv(errs))
+        fileno = getattr(sock, "fileno", None)
+        if fileno and fileno()==-1:
+            log("socket is closed, ignoring: %s", csv(errs))
+        else:
+            log.warn("Warning: failed to query %s", csv(errs))
+            log.warn(" on %s", sock)
     return opts
 
 
-class SSLPeekFile(object):
+class SocketPeekFile:
     def __init__(self, fileobj, peeked, update_peek):
         self.fileobj = fileobj
-        self.peeked = peeked
+        self.peeked : bytes = peeked
         self.update_peek = update_peek
 
     def __getattr__(self, attr):
@@ -480,7 +514,7 @@ class SSLPeekFile(object):
             return self.readline
         return getattr(self.fileobj, attr)
 
-    def readline(self, limit=-1):
+    def readline(self, limit:int=-1):
         if self.peeked:
             newline = self.peeked.find(b"\n")
             peeked = self.peeked
@@ -506,10 +540,11 @@ class SSLPeekFile(object):
             return peeked[:read]
         return self.fileobj.readline(limit)
 
-class SSLSocketWrapper(object):
-    def __init__(self, sock):
+
+class SocketPeekWrapper:
+    def __init__(self, sock, peeked=b""):
         self.socket = sock
-        self.peeked = b""
+        self.peeked = peeked
 
     def __getattr__(self, attr):
         if attr=="makefile":
@@ -521,13 +556,13 @@ class SSLSocketWrapper(object):
     def makefile(self, mode, bufsize=None):
         fileobj = self.socket.makefile(mode, bufsize)
         if self.peeked and mode and mode.startswith("r"):
-            return SSLPeekFile(fileobj, self.peeked, self._update_peek)
+            return SocketPeekFile(fileobj, self.peeked, self._update_peek)
         return fileobj
 
     def _update_peek(self, peeked):
         self.peeked = peeked
 
-    def recv(self, bufsize, flags=0):
+    def recv(self, bufsize, flags=0) -> bytes:
         if flags & socket.MSG_PEEK:
             l = len(self.peeked)
             if l>=bufsize:
@@ -549,7 +584,7 @@ class SSLSocketWrapper(object):
 class SSLSocketConnection(SocketConnection):
     SSL_TIMEOUT_MESSAGES = ("The read operation timed out", "The write operation timed out")
 
-    def can_retry(self, e):
+    def can_retry(self, e) -> Union[bool,str]:
         if getattr(e, "library", None)=="SSL":
             reason = getattr(e, "reason", None)
             if reason in ("WRONG_VERSION_NUMBER", "UNEXPECTED_RECORD"):
@@ -560,14 +595,10 @@ class SSLSocketConnection(SocketConnection):
         code = getattr(e, "code", None)
         if code in SSLSocketConnection.SSL_TIMEOUT_MESSAGES:
             return True
-        return SocketConnection.can_retry(self, e)
+        return super().can_retry(e)
 
-    def enable_peek(self):
-        assert not isinstance(self._socket, SSLSocketWrapper)
-        self._socket = SSLSocketWrapper(self._socket)
-
-    def get_info(self):
-        i = SocketConnection.get_info(self)
+    def get_info(self) -> Dict[str,Any]:
+        i = super().get_info()
         i["ssl"] = True
         for k,fn in {
                      "compression"      : "compression",
@@ -592,14 +623,17 @@ class SSLSocketConnection(SocketConnection):
         return i
 
 
-def set_socket_timeout(conn, timeout=None):
+def set_socket_timeout(conn, timeout=None) -> None:
     #FIXME: this is ugly, but less intrusive than the alternative?
-    log("set_socket_timeout(%s, %s)", conn, timeout)
     if isinstance(conn, SocketConnection):
+        sock = conn._socket
+        log("set_socket_timeout(%s, %s) applied to %s", conn, timeout, sock)
         conn._socket.settimeout(timeout)
+    else:
+        log("set_socket_timeout(%s, %s) ignored for %s", conn, timeout, type(conn))
 
 
-def log_new_connection(conn, socket_info=""):
+def log_new_connection(conn, socket_info:str="") -> None:
     """ logs the new connection message """
     sock = conn._socket
     address = conn.remote
@@ -615,17 +649,15 @@ def log_new_connection(conn, socket_info=""):
         sockname = ""
     log("log_new_connection(%s, %s) type=%s, sock=%s, sockname=%s, address=%s, peername=%s",
         conn, socket_info, type(conn), sock, sockname, address, peername)
+    log.info("New %s connection received", socktype)
     if peername:
         frominfo = pretty_socket(peername)
-        log.info("New %s connection received", socktype)
         log.info(" from '%s'", pretty_socket(frominfo))
         if socket_info:
             log.info(" on '%s'", pretty_socket(socket_info))
-    elif socktype=="unix-domain":
+    elif socktype=="socket":
         frominfo = sockname
-        log.info("New %s connection received", socktype)
         log.info(" on '%s'", frominfo)
     else:
-        log.info("New %s connection received", socktype)
         if socket_info:
             log.info(" on %s", pretty_socket(socket_info))
