@@ -1,13 +1,16 @@
 # This file is part of Xpra.
-# Copyright (C) 2011-2018 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2011-2023 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 import os
+import sys
 from ctypes import c_ubyte, c_char, c_uint32
+from typing import Tuple, Optional, Any, Callable
 
 from xpra.util import roundup
-from xpra.os_util import memoryview_to_bytes, shellsub, get_group_id, get_groups, WIN32, POSIX
+from xpra.common import noop
+from xpra.os_util import memoryview_to_bytes, shellsub, get_group_id, WIN32, POSIX
 from xpra.scripts.config import FALSE_OPTIONS, TRUE_OPTIONS
 from xpra.simple_stats import std_unit
 from xpra.log import Logger
@@ -21,18 +24,19 @@ MMAP_GROUP = os.environ.get("XPRA_MMAP_GROUP", "xpra")
 Utility functions for communicating via mmap
 """
 
-def get_socket_group(socket_filename):
+def get_socket_group(socket_filename) -> int:
     if isinstance(socket_filename, str) and os.path.exists(socket_filename):
         s = os.stat(socket_filename)
         return s.st_gid
-    log.warn("Warning: missing valid socket filename to set mmap group")
+    log.warn(f"Warning: missing valid socket filename {socket_filename!r} to set mmap group")
     return -1
 
-def xpra_group():
+def xpra_group() -> int:
     if POSIX:
         try:
             groups = os.getgroups()
             group_id = get_group_id(MMAP_GROUP)
+            log("xpra_group() group(%s)=%s, groups=%s", MMAP_GROUP, group_id, groups)
             if group_id and group_id in groups:
                 return group_id
         except Exception:
@@ -40,22 +44,24 @@ def xpra_group():
     return 0
 
 
-def init_client_mmap(mmap_group=None, socket_filename=None, size=128*1024*1024, filename=None):
+def init_client_mmap(mmap_group=None, socket_filename:str="", size:int=128*1024*1024, filename:str="") -> Tuple[bool, bool, Any, int, Any, str]:
     """
-        Initializes an mmap area, writes the token in it and returns:
+        Initializes a mmap area, writes the token in it and returns:
             (success flag, mmap_area, mmap_size, temp_file, mmap_filename)
         The caller must keep hold of temp_file to ensure it does not get deleted!
         This is used by the client.
     """
     def rerr():
-        return False, False, None, 0, None, None
+        return False, False, None, 0, None, ""
     log("init_mmap%s", (mmap_group, socket_filename, size, filename))
     mmap_filename = filename
     mmap_temp_file = None
     delete = True
-    def validate_size(size):
-        assert size>=64*1024*1024, "mmap size is too small: %sB (minimum is 64MB)" % std_unit(size)
-        assert size<=4*1024*1024*1024, "mmap is too big: %sB (maximum is 4GB)" % std_unit(size)
+    def validate_size(size : int):
+        if size<64*1024*1024:
+            raise ValueError("mmap size is too small: %sB (minimum is 64MB)" % std_unit(size))
+        if size>16*1024*1024*1024:
+            raise ValueError("mmap is too big: %sB (maximum is 4GB)" % std_unit(size))
     try:
         import mmap
         unit = max(4096, mmap.PAGESIZE)
@@ -82,17 +88,14 @@ def init_client_mmap(mmap_group=None, socket_filename=None, size=128*1024*1024,
                     log.info("Using existing mmap file '%s': %sMB", mmap_filename, mmap_size//1024//1024)
                 else:
                     validate_size(mmap_size)
-                    import errno
                     flags = os.O_CREAT | os.O_EXCL | os.O_RDWR
                     try:
                         fd = os.open(filename, flags)
                         mmap_temp_file = None   #os.fdopen(fd, 'w')
                         mmap_filename = filename
-                    except OSError as e:
-                        if e.errno == errno.EEXIST:
-                            log.error("Error: the mmap file '%s' already exists", filename)
-                            return rerr()
-                        raise
+                    except FileExistsError:
+                        log.error("Error: the mmap file '%s' already exists", filename)
+                        return rerr()
             else:
                 validate_size(mmap_size)
                 import tempfile
@@ -100,31 +103,30 @@ def init_client_mmap(mmap_group=None, socket_filename=None, size=128*1024*1024,
                 mmap_dir = get_mmap_dir()
                 subs = os.environ.copy()
                 subs.update({
-                    "UID"               : os.getuid(),
-                    "GID"               : os.getgid(),
-                    "PID"               : os.getpid(),
+                    "UID"               : str(os.getuid()),
+                    "GID"               : str(os.getgid()),
+                    "PID"               : str(os.getpid()),
                     })
                 mmap_dir = shellsub(mmap_dir, subs)
                 if mmap_dir and not os.path.exists(mmap_dir):
                     os.mkdir(mmap_dir, 0o700)
                 if not mmap_dir or not os.path.exists(mmap_dir):
-                    raise Exception("mmap directory %s does not exist!" % mmap_dir)
+                    raise RuntimeError("mmap directory %s does not exist!" % mmap_dir)
                 #create the mmap file, the mkstemp that is called via NamedTemporaryFile ensures
                 #that the file is readable and writable only by the creating user ID
                 try:
                     temp = tempfile.NamedTemporaryFile(prefix="xpra.", suffix=".mmap", dir=mmap_dir)
                 except OSError as e:
-                    log.error("Error: cannot create mmap file:")
-                    log.error(" %s", e)
+                    log.error("Error: cannot create mmap temporary file:")
+                    log.estr(e)
                     return rerr()
-                #keep a reference to it so it does not disappear!
+                #keep a reference to it, so it does not disappear!
                 mmap_temp_file = temp
                 mmap_filename = temp.name
                 fd = temp.file.fileno()
             #set the group permissions and gid if the mmap-group option is specified
             mmap_group = (mmap_group or "")
             if POSIX and mmap_group and mmap_group not in FALSE_OPTIONS:
-                group_id = None
                 if mmap_group=="SOCKET":
                     group_id = get_socket_group(socket_filename)
                 elif mmap_group.lower()=="auto":
@@ -144,7 +146,7 @@ def init_client_mmap(mmap_group=None, socket_filename=None, size=128*1024*1024,
                     except OSError as e:
                         log("fchown(%i, %i, %i) on %s", fd, -1, group_id, mmap_filename, exc_info=True)
                         log.error("Error: failed to change group ownership of mmap file to '%s':", mmap_group)
-                        log.error(" %s", e)
+                        log.estr(e)
                     from stat import S_IRUSR,S_IWUSR,S_IRGRP,S_IWGRP
                     os.fchmod(fd, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP)
             log("using mmap file %s, fd=%s, size=%s", mmap_filename, fd, mmap_size)
@@ -156,7 +158,7 @@ def init_client_mmap(mmap_group=None, socket_filename=None, size=128*1024*1024,
     except Exception as e:
         log("failed to setup mmap: %s", e, exc_info=True)
         log.error("Error: mmap setup failed:")
-        log.error(" %s", e)
+        log.estr(e)
         if delete:
             if mmap_temp_file:
                 try:
@@ -168,19 +170,18 @@ def init_client_mmap(mmap_group=None, socket_filename=None, size=128*1024*1024,
                 clean_mmap(mmap_filename)
         return rerr()
 
-def clean_mmap(mmap_filename):
+def clean_mmap(mmap_filename:str) -> None:
     log("clean_mmap(%s)", mmap_filename)
     if mmap_filename and os.path.exists(mmap_filename):
         try:
             os.unlink(mmap_filename)
         except OSError as e:
             log.error("Error: failed to remove the mmap file '%s':", mmap_filename)
-            log.error(" %s", e)
+            log.estr(e)
 
-DEFAULT_TOKEN_INDEX = 512
-DEFAULT_TOKEN_BYTES = 128
+DEFAULT_TOKEN_BYTES : int = 128
 
-def write_mmap_token(mmap_area, token, index=DEFAULT_TOKEN_INDEX, count=DEFAULT_TOKEN_BYTES):
+def write_mmap_token(mmap_area, token, index:int, count:int=DEFAULT_TOKEN_BYTES) -> None:
     assert count>0
     #write the token one byte at a time - no endianness
     log("write_mmap_token(%s, %#x, %#x, %#x)", mmap_area, token, index, count)
@@ -191,7 +192,7 @@ def write_mmap_token(mmap_area, token, index=DEFAULT_TOKEN_INDEX, count=DEFAULT_
         v = v>>8
     assert v==0, "token value is too big"
 
-def read_mmap_token(mmap_area, index=DEFAULT_TOKEN_INDEX, count=DEFAULT_TOKEN_BYTES):
+def read_mmap_token(mmap_area, index:int, count:int=DEFAULT_TOKEN_BYTES) -> int:
     assert count>0
     v = 0
     for i in range(0, count):
@@ -202,45 +203,44 @@ def read_mmap_token(mmap_area, index=DEFAULT_TOKEN_INDEX, count=DEFAULT_TOKEN_BY
     return v
 
 
-def init_server_mmap(mmap_filename, mmap_size=0):
+def init_server_mmap(mmap_filename:str, mmap_size:int=0) -> Tuple[Optional[Any],int]:
     """
         Reads the mmap file provided by the client
         and verifies the token if supplied.
         Returns the mmap object and its size: (mmap, size)
     """
-    if not WIN32:
-        try:
-            f = open(mmap_filename, "r+b")
-        except Exception as e:
-            log.error("Error: cannot access mmap file '%s':", mmap_filename)
-            log.error("  %s", e)
-            log.error(" see mmap-group option?")
-            return None, 0
-
     mmap_area = None
     try:
         import mmap
         if not WIN32:
+            try:
+                f = open(mmap_filename, "r+b")
+            except Exception as e:
+                log.error(f"Error: cannot access mmap file {mmap_filename!r}:")
+                log.estr(e)
+                log.error(" see mmap-group option?")
+                return None, 0
             actual_mmap_size = os.path.getsize(mmap_filename)
             if mmap_size and actual_mmap_size!=mmap_size:
                 log.warn("Warning: expected mmap file '%s' of size %i but got %i",
                          mmap_filename, mmap_size, actual_mmap_size)
             mmap_area = mmap.mmap(f.fileno(), mmap_size)
-        else:
-            if mmap_size==0:
-                log.error("Error: client did not supply the mmap area size")
-                log.error(" try updating your client version?")
-            mmap_area = mmap.mmap(0, mmap_size, mmap_filename)
-            actual_mmap_size = mmap_size
-        f.close()
-        return mmap_area, actual_mmap_size
-    except Exception as e:
-        log.error("cannot use mmap file '%s': %s", mmap_filename, e, exc_info=True)
+            f.close()
+            return mmap_area, actual_mmap_size
+        assert sys.platform=="win32"
+        if mmap_size==0:
+            log.error("Error: client did not supply the mmap area size")
+            log.error(" try updating your client version?")
+            return None, 0
+        mmap_area = mmap.mmap(0, mmap_size, mmap_filename)
+        return mmap_area, mmap_size
+    except Exception:
+        log.error("Error: cannot use mmap file '%s'", mmap_filename, exc_info=True)
         if mmap_area:
             mmap_area.close()
         return None, 0
 
-def int_from_buffer(mmap_area, pos):
+def int_from_buffer(mmap_area, pos:int) -> c_uint32:
     return c_uint32.from_buffer(mmap_area, pos)      #@UndefinedVariable
 
 
@@ -251,23 +251,24 @@ def mmap_read(mmap_area, *descr_data):
         Reads data from the mmap_area as written by 'mmap_write'.
         The descr_data is the list of mmap chunks used.
     """
-    data_start = int_from_buffer(mmap_area, 0)
+    data_start : c_uint32 = int_from_buffer(mmap_area, 0)
     if len(descr_data)==1:
         #construct an array directly from the mmap zone:
         offset, length = descr_data[0]
         arraytype = c_char * length
-        data_start.value = offset+length
-        return arraytype.from_buffer(mmap_area, offset)
+        def free_mem(*_args):
+            data_start.value = offset + length
+        return arraytype.from_buffer(mmap_area, offset), free_mem
     #re-construct the buffer from discontiguous chunks:
     data = []
     for offset, length in descr_data:
         mmap_area.seek(offset)
         data.append(mmap_area.read(length))
         data_start.value = offset+length
-    return b"".join(data)
+    return b"".join(data), noop
 
 
-def mmap_write(mmap_area, mmap_size, data):
+def mmap_write(mmap_area, mmap_size:int, data):
     """
         Sends 'data' to the client via the mmap shared memory region,
         returns the chunks of the mmap area used (or None if it failed)
@@ -282,8 +283,8 @@ def mmap_write(mmap_area, mmap_size, data):
     # '+' is for data we have written
     # '*' is for data we have just written in this call
     # E and S show the location pointed to by data_start/data_end
-    mmap_data_start = int_from_buffer(mmap_area, 0)
-    mmap_data_end = int_from_buffer(mmap_area, 4)
+    mmap_data_start : c_uint32 = int_from_buffer(mmap_area, 0)
+    mmap_data_end : c_uint32 = int_from_buffer(mmap_area, 4)
     start = max(8, mmap_data_start.value)
     end = max(8, mmap_data_end.value)
     l = len(data)
