@@ -1,106 +1,64 @@
 # This file is part of Xpra.
-# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2019-2023 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 import os
 import struct
+from typing import List, ByteString, Callable
 
-from xpra.net.websockets.header import encode_hybi_header, decode_hybi
-from xpra.net.protocol import Protocol
+from xpra.net.websockets.mask import hybi_mask     #@UnresolvedImport
+from xpra.net.websockets.header import encode_hybi_header, decode_hybi, close_packet
+from xpra.net.websockets.common import (
+    OPCODES,
+    OPCODE_BINARY, OPCODE_CONTINUE, OPCODE_TEXT, OPCODE_CLOSE, OPCODE_PING, OPCODE_PONG,
+    )
+from xpra.net.protocol.socket_handler import SocketProtocol
 from xpra.util import first_time, envbool
-from xpra.os_util import memoryview_to_bytes
+from xpra.os_util import memoryview_to_bytes, hexstr
 from xpra.log import Logger
 
 log = Logger("websocket")
 
-OPCODE_CONTINUE = 0
-OPCODE_TEXT     = 1
-OPCODE_BINARY   = 2
-OPCODE_CLOSE    = 8
-OPCODE_PING     = 9
-OPCODE_PONG     = 10
-OPCODES = {
-    OPCODE_CONTINUE : "continuation frame",
-    OPCODE_TEXT     : "text frame",
-    OPCODE_BINARY   : "binary frame",
-    OPCODE_CLOSE    : "connection close",
-    OPCODE_PING     : "ping",
-    OPCODE_PONG     : "pong",
-    }
-
-#default to legacy mode until we parse the remote caps:
-#(this can be removed in the future once all html5 clients have been updated)
-LEGACY_FRAME_PER_CHUNK = envbool("XPRA_WEBSOCKET_LEGACY", None)
 MASK = envbool("XPRA_WEBSOCKET_MASK", False)
 
 
-class WebSocketProtocol(Protocol):
-
-    STATE_FIELDS = tuple(list(Protocol.STATE_FIELDS)+["legacy_frame_per_chunk"])
+class WebSocketProtocol(SocketProtocol):
 
     TYPE = "websocket"
 
     def __init__(self, *args, **kwargs):
-        Protocol.__init__(self, *args, **kwargs)
-        self.ws_data = b""
-        self.ws_payload = []
-        self.ws_payload_opcode = 0
-        self.ws_mask = MASK
+        super().__init__(*args, **kwargs)
+        self.ws_data : ByteString = b""
+        self.ws_payload : List[ByteString] = []
+        self.ws_payload_opcode : int = 0
+        self.ws_mask : bool = MASK
         self._process_read = self.parse_ws_frame
-        self.legacy_frame_per_chunk = LEGACY_FRAME_PER_CHUNK in (None, True)
-        if self.legacy_frame_per_chunk:
-            self.make_chunk_header = self.make_wschunk_header
-        else:
-            self.make_chunk_header = self.make_xpra_header
-            self.make_frame_header = self.make_wsframe_header
+        self.make_chunk_header : Callable = self.make_xpra_header
+        self.make_frame_header : Callable = self.make_wsframe_header
 
     def __repr__(self):
-        return "WebSocket(%s)" % self._conn
+        return f"WebSocket({self._conn})"
 
-    def close(self):
-        Protocol.close(self)
+    def close(self, message=None) -> None:
+        if self._closed:
+            return
+        self.send_ws_close(reason=message)
+        super().close(message)
         self.ws_data = b""
         self.ws_payload = []
 
-
-    def get_info(self, alias_info=True):
-        info = Protocol.get_info(self, alias_info)
-        info["legacy-frames"] = self.legacy_frame_per_chunk
-        return info
+    def send_ws_close(self, code:int=1000, reason:str="closing") -> None:
+        data = close_packet(code, reason)
+        self.flush_then_close(None, data)
 
 
-    def parse_remote_caps(self, caps):
-        Protocol.parse_remote_caps(self, caps)
-        if LEGACY_FRAME_PER_CHUNK is None:
-            may_have_bug = caps.strget("client_type", "")=="HTML5"
-            self.legacy_frame_per_chunk = not caps.boolget("websocket.multi-packet", not may_have_bug)
-            log("parse_remote_caps() may_have_bug=%s, legacy_frame_per_chunk=%s",
-                may_have_bug, self.legacy_frame_per_chunk)
-        if self.legacy_frame_per_chunk:
-            log.warn("Warning: using slower legacy websocket frames")
-            log.warn(" the other end is probably out of date")
-            #websocker header for every chunk:
-            self.make_chunk_header = self.make_wschunk_header
-            #no frame header:
-            self.make_frame_header = self.noframe_header
-        else:
-            #just the regular xpra header for each chunk:
-            self.make_chunk_header = self.make_xpra_header
-            #and one websocket header for all the chunks:
-            self.make_frame_header = self.make_wsframe_header
-
-    def make_wschunk_header(self, packet_type, proto_flags, level, index, payload_size, total_size):
-        header = Protocol.make_xpra_header(self, packet_type, proto_flags, level, index, payload_size, total_size)
-        log("make_wschunk_header(%s)", (packet_type, proto_flags, level, index, payload_size))
-        return encode_hybi_header(OPCODE_BINARY, total_size+len(header))+header
-
-    def make_wsframe_header(self, packet_type, items):
+    def make_wsframe_header(self, packet_type, items) -> ByteString:
         payload_len = sum(len(item) for item in items)
-        log("make_wsframe_header(%s, %i items) %i bytes", packet_type, len(items), payload_len)
         header = encode_hybi_header(OPCODE_BINARY, payload_len, self.ws_mask)
+        log("make_wsframe_header(%s, %i items) %i bytes, ws_mask=%s, header=0x%s (%i bytes)",
+            packet_type, len(items), payload_len, self.ws_mask, hexstr(header), len(header))
         if self.ws_mask:
-            from xpra.codecs.xor.cyxor import hybi_mask     #@UnresolvedImport
             mask = os.urandom(4)
             #now mask all the items:
             for i, item in enumerate(items):
@@ -108,20 +66,22 @@ class WebSocketProtocol(Protocol):
             return header+mask
         return header
 
-    def parse_ws_frame(self, buf):
+    def parse_ws_frame(self, buf:ByteString) -> None:
         if not buf:
             self._read_queue_put(buf)
             return
         if self.ws_data:
-            ws_data = self.ws_data+buf
+            ws_data = b"".join((self.ws_data, buf))
             self.ws_data = b""
         else:
             ws_data = buf
+        while self.input_packetcount==0 and ws_data.startswith(b"\r\n"):
+            ws_data = ws_data[2:]
         log("parse_ws_frame(%i bytes) total buffer is %i bytes", len(buf), len(ws_data))
         while ws_data and not self._closed:
             parsed = decode_hybi(ws_data)
             if parsed is None:
-                log("parse_ws_frame(%i bytes) not enough data", len(ws_data))
+                log("parse_ws_frame(%i bytes) not enough data: %r", len(ws_data), ws_data)
                 #not enough data to get a full websocket frame,
                 #save it for later:
                 self.ws_data = ws_data
@@ -143,19 +103,26 @@ class WebSocketProtocol(Protocol):
                 self.ws_payload_opcode = 0
             else:
                 if self.ws_payload and self.ws_payload_opcode:
-                    raise Exception("expected a continuation frame not %s" % OPCODES.get(opcode, opcode))
+                    op = OPCODES.get(opcode, opcode)
+                    raise ValueError(f"expected a continuation frame not {op}")
                 full_payload = payload
                 if not fin:
                     if opcode not in (OPCODE_BINARY, OPCODE_TEXT):
-                        raise Exception("cannot handle fragmented '%s' frames" % OPCODES.get(opcode, opcode))
+                        op = OPCODES.get(opcode, opcode)
+                        log(f"invalid opcode {opcode} from {buf!r}")
+                        log(f"parsed as {parsed}")
+                        raise RuntimeError(f"cannot handle fragmented {op} frames")
                     #fragmented, keep this payload for later
                     self.ws_payload_opcode = opcode
                     self.ws_payload.append(payload)
                     continue
             if opcode==OPCODE_BINARY:
+                if not full_payload:
+                    log.warn("Warning: empty websocket binary payload")
+                    continue
                 self._read_queue_put(full_payload)
             elif opcode==OPCODE_TEXT:
-                if first_time("ws-text-frame-from-%s" % self._conn):
+                if first_time(f"ws-text-frame-from-{self._conn}"):
                     log.warn("Warning: handling text websocket frame as binary")
                 self._read_queue_put(full_payload)
             elif opcode==OPCODE_CLOSE:
@@ -165,20 +132,20 @@ class WebSocketProtocol(Protocol):
             elif opcode==OPCODE_PONG:
                 self._process_ws_pong(full_payload)
             else:
-                log.warn("Warning unhandled websocket opcode '%s'", OPCODES.get(opcode, "%#x" % opcode))
+                log.warn("Warning unhandled websocket opcode '%s'", OPCODES.get(opcode, f"{opcode:x}"))
                 log("payload=%r", payload)
 
-    def _process_ws_ping(self, payload):
+    def _process_ws_ping(self, payload:ByteString) -> None:
         log("_process_ws_ping(%r)", payload)
         item = encode_hybi_header(OPCODE_PONG, len(payload)) + memoryview_to_bytes(payload)
         items = (item, )
         with self._write_lock:
             self.raw_write("ws-ping", items)
 
-    def _process_ws_pong(self, payload):
+    def _process_ws_pong(self, payload:ByteString) -> None:
         log("_process_ws_pong(%r)", payload)
 
-    def _process_ws_close(self, payload):
+    def _process_ws_close(self, payload:ByteString) -> None:
         log("_process_ws_close(%r)", payload)
         if len(payload)<2:
             self._connection_lost("unknown reason")
@@ -188,4 +155,4 @@ class WebSocketProtocol(Protocol):
             reason = memoryview_to_bytes(payload[2:]).decode("utf8")
         except UnicodeDecodeError:
             reason = str(payload[2:])
-        self._connection_lost("code %s: %s" % (code, reason))
+        self._connection_lost(f"code {code}: {reason}")
