@@ -268,8 +268,6 @@ def get_int_uuid() -> int:
 def get_machine_id() -> str:
     """
         Try to get uuid string which uniquely identifies this machine.
-        Warning: only works on posix!
-        (which is ok since we only used it on posix at present)
     """
     v = ""
     if POSIX:
@@ -300,6 +298,8 @@ def get_user_uuid() -> str:
         uupdate(str(os.getuid()))
         uupdate("/")
         uupdate(str(os.getgid()))
+        uupdate("/")
+        uupdate(os.environ.get("DISPLAY", "") or os.environ.get("WAYLAND_DISPLAY", ""))
     uupdate(os.path.expanduser("~/"))
     return u.hexdigest()
 
