@@ -1,40 +1,43 @@
 # This file is part of Xpra.
-# Copyright (C) 2011-2018 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2011-2023 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
+from typing import List, Optional, Callable
+
+import gi
+gi.require_version('Gtk', '3.0')  # @UndefinedVariable
+from gi.repository import GLib, Gtk  # @UnresolvedImport
+
 from xpra.util import envbool, csv
 from xpra.gtk_common.gtk_util import scaled_image
 from xpra.gtk_common.about import about
-from xpra.client.gtk_base.gtk_tray_menu_base import (
-    GTKTrayMenuBase, populate_encodingsmenu,
+from xpra.client.gtk3.gtk_tray_menu_base import (
+    GTKTrayMenuBase,
     CLIPBOARD_LABEL_TO_NAME, CLIPBOARD_NAME_TO_LABEL, CLIPBOARD_LABELS,
     CLIPBOARD_DIRECTION_LABELS, CLIPBOARD_DIRECTION_NAME_TO_LABEL,
     SHOW_UPLOAD, SHOW_VERSION_CHECK, RUNCOMMAND_MENU, SHOW_SERVER_COMMANDS, SHOW_SHUTDOWN,
+    SHOW_QR,
     )
 from xpra.platform.paths import get_icon
-from xpra.platform.darwin.gui import get_OSXApplication
-from xpra.client import mixin_features
-from xpra.gtk_common.gobject_compat import import_gtk, import_glib
+from xpra.platform.darwin import get_OSXApplication
+from xpra.client.gui import mixin_features
 from xpra.log import Logger
 
 log = Logger("osx", "tray", "menu")
 clipboardlog = Logger("osx", "menu", "clipboard")
 
-gtk = import_gtk()
-glib = import_glib()
-
-
 #control which menus are shown in the OSX global menu:
-SHOW_FEATURES_MENU = True
-SHOW_SOUND_MENU = True
-SHOW_ENCODINGS_MENU = True
-SHOW_ACTIONS_MENU = True
-SHOW_INFO_MENU = True
-SHOW_CLIPBOARD_MENU = True
-SHOW_SERVER_MENU = True
+DEFAULT = envbool("XPRA_OSX_SHOW_MENU_DEFAULT", True)
+SHOW_FEATURES_MENU = DEFAULT
+SHOW_SOUND_MENU = DEFAULT
+SHOW_ENCODINGS_MENU = DEFAULT
+SHOW_ACTIONS_MENU = DEFAULT
+SHOW_INFO_MENU = DEFAULT
+SHOW_CLIPBOARD_MENU = DEFAULT
+SHOW_SERVER_MENU = DEFAULT
 
-SHOW_ABOUT_XPRA = True
+SHOW_ABOUT_XPRA = DEFAULT
 
 SINGLE_MENU = envbool("XPRA_OSX_SINGLE_MENU", False)
 USE_WINDOW_MENU = envbool("XPRA_OSX_USE_WINDOW_MENU", True)
@@ -55,39 +58,42 @@ def getOSXMenuHelper(client=None):
 
 class OSXMenuHelper(GTKTrayMenuBase):
     """
-    we have to do this stuff here so we can
-    re-use the same instance,
+    we have to do this stuff here,
+    so we can re-use the same instance,
     and change the callbacks if needed.
     (that way, the launcher and the client can both change the menus)
     """
 
     def __init__(self, client=None):
-        GTKTrayMenuBase.__init__(self, client)
+        super().__init__(client)
         log("OSXMenuHelper(%s)", client)
         self.menu_bar = None
         self.hidden_window = None
         self.menus = {}             #all the "top-level" menu items we manage
         self.app_menus = {}         #the ones added to the app_menu via insert_app_menu_item (which cannot be removed!)
+        self.dock_menu = None
+        self.window_menu = None
         self.window_menu_item = None
+        self.numlock_menuitem = None
         self.full = False
         self.set_client(client)
         self._clipboard_change_pending = False
 
-    def set_client(self, client):
+    def set_client(self, client) -> None:
         self.client = client
         #if we call add_about before the main loop is ready,
         #things don't work...
         if client and SHOW_ABOUT_XPRA:
             client.after_handshake(self.add_about)
 
-    def show_menu(self, button, time):
+    def show_menu(self, button:int, time) -> None:
         #does not mean anything on OSX since the menu is controlled by the OS
         pass
 
     def build(self):
         log("OSXMenuHelper.build()")
         if self.menu_bar is None:
-            self.menu_bar = gtk.MenuBar()
+            self.menu_bar = Gtk.MenuBar()
             self.menu_bar.show_all()
         return self.menu_bar
 
@@ -98,7 +104,7 @@ class OSXMenuHelper(GTKTrayMenuBase):
         self.remove_all_menus()
         return self.menu_bar
 
-    def remove_all_menus(self):
+    def remove_all_menus(self) -> None:
         log("OSXMenuHelper.remove_all_menus()")
         if self.menu_bar:
             for x in self.menus.values():
@@ -109,7 +115,7 @@ class OSXMenuHelper(GTKTrayMenuBase):
         self.full = False
 
 
-    def add_top_level_menu(self, label, submenu):
+    def add_top_level_menu(self, label:str, submenu) -> None:
         """ Adds the item to the app-menu or to the top bar,
             but only if it has not been added yet.
             (depending on the SINGLE_MENU flag)
@@ -120,7 +126,7 @@ class OSXMenuHelper(GTKTrayMenuBase):
         else:
             self.add_to_menu_bar(label, submenu)
 
-    def add_to_app_menu(self, label, submenu):
+    def add_to_app_menu(self, label:str, submenu) -> None:
         item = self.app_menus.get(label)
         if item:
             log("application menu already has a '%s' entry", label)
@@ -129,7 +135,7 @@ class OSXMenuHelper(GTKTrayMenuBase):
             item.show_all()
         else:
             if label.startswith(SEPARATOR):
-                item = gtk.SeparatorMenuItem()
+                item = Gtk.SeparatorMenuItem()
             else:
                 item = self.menuitem(label)
                 item.set_submenu(submenu)
@@ -152,13 +158,13 @@ class OSXMenuHelper(GTKTrayMenuBase):
             item.set_submenu(submenu)
             item.show_all()
 
-    def get_menu(self, label):
+    def get_menu(self, label:str):
         if SINGLE_MENU:
             return self.app_menus.get(label)
         return self.menus.get(label)
 
 
-    def add_about(self):
+    def add_about(self) -> None:
         if "About" in self.app_menus:
             return
         item = self.menuitem("About", cb=about)
@@ -168,7 +174,7 @@ class OSXMenuHelper(GTKTrayMenuBase):
         self.app_menus["About"] = item
 
 
-    def add_full_menu(self):
+    def add_full_menu(self) -> None:
         log("OSXMenuHelper.add_full_menu()")
         if self.full or not self.client:
             return
@@ -183,87 +189,89 @@ class OSXMenuHelper(GTKTrayMenuBase):
                 macapp = get_OSXApplication()
                 macapp.set_window_menu(self.window_menu_item)
 
-    def get_extra_menus(self):
+    def get_extra_menus(self) -> List:
         menus = []
+        def add(menu, item):
+            if item:
+                menu.add(item)
         if SHOW_INFO_MENU:
             info_menu = self.make_menu()
-            info_menu.append(self.make_sessioninfomenuitem())
-            if SHOW_VERSION_CHECK:
-                info_menu.append(self.make_updatecheckmenuitem())
-            info_menu.append(self.make_bugreportmenuitem())
             menus.append(("Info", info_menu))
+            add(info_menu, self.make_sessioninfomenuitem())
+            if SHOW_QR:
+                add(info_menu, self.make_qrmenuitem())
+            if SHOW_VERSION_CHECK:
+                add(info_menu, self.make_updatecheckmenuitem())
+            add(info_menu, self.make_bugreportmenuitem())
         if SHOW_FEATURES_MENU:
             features_menu = self.make_menu()
             menus.append(("Features", features_menu))
             self.append_featuresmenuitems(features_menu)
             if mixin_features.windows:
-                features_menu.add(self.make_swapkeysmenuitem())
-                features_menu.add(self.make_invertmousewheelmenuitem())
-                features_menu.add(self.make_numlockmenuitem())
-                features_menu.add(self.make_scalingmenuitem())
+                add(features_menu, self.make_swapkeysmenuitem())
+                add(features_menu, self.make_invertmousewheelmenuitem())
+                add(features_menu, self.make_numlockmenuitem())
+                add(features_menu, self.make_scalingmenuitem())
         if mixin_features.clipboard and SHOW_CLIPBOARD_MENU:
             clipboard_menu = self.make_menu()
             menus.append(("Clipboard", clipboard_menu))
             for label in CLIPBOARD_LABELS:
-                clipboard_menu.add(self.make_clipboard_submenuitem(label, self._remote_clipboard_changed))
-            clipboard_menu.add(gtk.SeparatorMenuItem())
+                add(clipboard_menu, self.make_clipboard_submenuitem(label, self._remote_clipboard_changed))
+            add(clipboard_menu, Gtk.SeparatorMenuItem())
             for label in CLIPBOARD_DIRECTION_LABELS:
-                clipboard_menu.add(self.make_clipboard_submenuitem(label, self._clipboard_direction_changed))
+                add(clipboard_menu, self.make_clipboard_submenuitem(label, self._clipboard_direction_changed))
             clipboard_menu.show_all()
-            self.client.after_handshake(self.set_clipboard_menu, clipboard_menu)
+            self.after_handshake(self.set_clipboard_menu, clipboard_menu)
         if mixin_features.audio and SHOW_SOUND_MENU:
-            sound_menu = self.make_menu()
+            audio_menu = self.make_menu()
             if self.client.speaker_allowed and self.client.speaker_codecs:
-                sound_menu.add(self.make_speakermenuitem())
+                add(audio_menu, self.make_speakermenuitem())
             if self.client.microphone_allowed and self.client.microphone_codecs:
-                sound_menu.add(self.make_microphonemenuitem())
-            menus.append(("Sound", sound_menu))
+                add(audio_menu, self.make_microphonemenuitem())
+            menus.append(("Audio", audio_menu))
         if mixin_features.windows and SHOW_ENCODINGS_MENU:
             encodings_menu = self.make_menu()
             def set_encodings_menu(*_args):
-                from xpra.codecs.codec_constants import PREFERED_ENCODING_ORDER
-                server_encodings = list(self.client.server_encodings)
-                encodings = [x for x in PREFERED_ENCODING_ORDER if x in self.client.get_encodings()]
-                if self.client.server_auto_video_encoding:
-                    encodings.insert(0, "auto")
-                    server_encodings.insert(0, "auto")
-                populate_encodingsmenu(encodings_menu, self.get_current_encoding, self.set_current_encoding, encodings, server_encodings)
-            self.client.after_handshake(set_encodings_menu)
+                client_encodings, server_encodings = self.get_encoding_options()
+                from xpra.client.gtk3.menu_helper import populate_encodingsmenu
+                populate_encodingsmenu(encodings_menu, self.get_current_encoding, self.set_current_encoding,
+                                       client_encodings, server_encodings)
+            self.after_handshake(set_encodings_menu)
             menus.append(("Encoding", encodings_menu))
         if mixin_features.windows and SHOW_ACTIONS_MENU:
             actions_menu = self.make_menu()
-            actions_menu.add(self.make_raisewindowsmenuitem())
-            actions_menu.add(self.make_minimizewindowsmenuitem())
-            actions_menu.add(self.make_refreshmenuitem())
-            actions_menu.add(self.make_reinitmenuitem())
+            add(actions_menu, self.make_raisewindowsmenuitem())
+            add(actions_menu, self.make_minimizewindowsmenuitem())
+            add(actions_menu, self.make_refreshmenuitem())
+            add(actions_menu, self.make_reinitmenuitem())
             self.window_menu = actions_menu
             menus.append(("Windows", actions_menu))
         if RUNCOMMAND_MENU or SHOW_SERVER_COMMANDS or SHOW_UPLOAD or SHOW_SHUTDOWN:
             server_menu = self.make_menu()
             if SHOW_SHUTDOWN:
-                server_menu.append(self.make_shutdownmenuitem())
+                add(server_menu, self.make_shutdownmenuitem())
             #set_sensitive(bool) does not work on OSX,
             #so we only add the menu item if it does something
             def add_ah(*_args):
                 if self.client.server_start_new_commands:
-                    server_menu.add(self.make_runcommandmenuitem())
+                    add(server_menu, self.make_runcommandmenuitem())
                 if SHOW_SERVER_COMMANDS and self.client.server_commands_info:
-                    server_menu.append(self.make_servercommandsmenuitem())
+                    add(server_menu, self.make_servercommandsmenuitem())
                 if SHOW_UPLOAD and self.client.remote_file_transfer:
-                    server_menu.add(self.make_uploadmenuitem())
-            self.client.after_handshake(add_ah)
+                    add(server_menu, self.make_uploadmenuitem())
+            self.after_handshake(add_ah)
             menus.append(("Server", server_menu))
         menus.append((SEPARATOR+"-EXTRAS", None))
         return menus
 
 
-    def _clipboard_direction_changed(self, item, label):
+    def _clipboard_direction_changed(self, item, label:str):
         clipboardlog("_clipboard_direction_changed(%s, %s) clipboard_change_pending=%s",
                      item, label, self._clipboard_change_pending)
         label = self.select_clipboard_menu_option(item, label, CLIPBOARD_DIRECTION_LABELS)
         self.do_clipboard_direction_changed(label or "")
 
-    def _remote_clipboard_changed(self, item, label):
+    def _remote_clipboard_changed(self, item, label:str):
         clipboardlog("_remote_clipboard_changed(%s, %s) clipboard_change_pending=%s",
                      item, label, self._clipboard_change_pending)
         #ensure this is the only clipboard label selected:
@@ -272,9 +280,9 @@ class OSXMenuHelper(GTKTrayMenuBase):
             return
         remote_clipboard = CLIPBOARD_LABEL_TO_NAME[label]
         clipboardlog("will select clipboard menu item with label=%s, for remote_clipboard=%s", label, remote_clipboard)
-        glib.timeout_add(0, self._do_clipboard_change, remote_clipboard)
+        GLib.timeout_add(0, self._do_clipboard_change, remote_clipboard)
 
-    def _do_clipboard_change(self, remote_clipboard):
+    def _do_clipboard_change(self, remote_clipboard:str):
         #why do we look it up again when we could just pass it in
         #to make_clipboard_submenuitem as an extra argument?
         #because gtk-osx would fall over itself, making a complete mess of the menus in the process
@@ -282,7 +290,7 @@ class OSXMenuHelper(GTKTrayMenuBase):
         self._clipboard_change_pending = False
         self.set_new_remote_clipboard(remote_clipboard)
 
-    def make_clipboard_submenuitem(self, label, cb=None):
+    def make_clipboard_submenuitem(self, label:str, cb:Optional[Callable]=None):
         clipboard_item = self.checkitem(label)
         clipboard_item.set_draw_as_radio(True)
         def clipboard_option_changed(item):
@@ -293,7 +301,7 @@ class OSXMenuHelper(GTKTrayMenuBase):
         clipboard_item.connect("toggled", clipboard_option_changed)
         return clipboard_item
 
-    def select_clipboard_menu_option(self, item=None, label=None, labels=()):
+    def select_clipboard_menu_option(self, item=None, label:str="", labels=()):
         #ensure that only the matching menu item is selected,
         #(can be specified as a menuitem object, or using its label)
         #all the other menu items whose labels are specified will be made inactive
@@ -341,12 +349,13 @@ class OSXMenuHelper(GTKTrayMenuBase):
         self.select_clipboard_menu_option(None, direction_label, CLIPBOARD_DIRECTION_LABELS)
 
 
-    #these methods are called by the superclass
-    #but we don't have a quality or speed menu, so override and ignore
+    # these methods are called by the superclass,
+    # but we don't have a quality or speed menu,
+    # so override and ignore
     def set_qualitymenu(self, *args):
-        pass
+        pass    #no quality menu on MacOS
     def set_speedmenu(self, *args):
-        pass
+        pass    #no speed menu on MacOS
 
     def _get_keyboard(self):
         if not self.client or not self.client.keyboard_helper:
@@ -355,26 +364,26 @@ class OSXMenuHelper(GTKTrayMenuBase):
 
     def make_swapkeysmenuitem(self):
         def swapkeys_toggled(*args):
-            v = self.swapkeys_menuitem.get_active()
+            v = swapkeys_menuitem.get_active()
             keyboard = self._get_keyboard()
             log("swapkeys_toggled(%s) keyboard=%s, swap keys enabled=%s", args, keyboard, v)
             if keyboard:
                 keyboard.swap_keys = v
-        self.swapkeys_menuitem = self.checkitem("Control/Command Key Swap", swapkeys_toggled)
+        swapkeys_menuitem = self.checkitem("Control/Command Key Swap", swapkeys_toggled)
         def set_swapkeys_menuitem(*args):
             keyboard = self._get_keyboard()
             if keyboard:
                 log("set_swapkeys_menuitem(%s) keyboard=%s, swap_keys=%s", args, keyboard, keyboard.swap_keys)
-                self.swapkeys_menuitem.set_active(keyboard.swap_keys)
+                swapkeys_menuitem.set_active(keyboard.swap_keys)
             else:
                 log("set_swapkeys_menuitem(%s) no keyboard!", args)
-                self.swapkeys_menuitem.set_sensitive(False)
-        self.client.after_handshake(set_swapkeys_menuitem)
-        return self.swapkeys_menuitem
+                swapkeys_menuitem.set_sensitive(False)
+        self.after_handshake(set_swapkeys_menuitem)
+        return swapkeys_menuitem
 
     def make_invertmousewheelmenuitem(self):
         def invert_toggled(*args):
-            v = self.mousewheel_menuitem.get_active()
+            v = mousewheel_menuitem.get_active()
             log("invert_toggled(%s) invert enabled=%s", args, v)
             if v:
                 self.client.wheel_map[4] = 5
@@ -382,9 +391,9 @@ class OSXMenuHelper(GTKTrayMenuBase):
             else:
                 self.client.wheel_map[4] = 4
                 self.client.wheel_map[5] = 5
-        self.mousewheel_menuitem = self.checkitem("Invert Mouse Wheel", invert_toggled)
-        self.mousewheel_menuitem.set_active(self.client.wheel_map.get(4)!=4)
-        return self.mousewheel_menuitem
+        mousewheel_menuitem = self.checkitem("Invert Mouse Wheel", invert_toggled)
+        mousewheel_menuitem.set_active(self.client.wheel_map.get(4)!=4)
+        return mousewheel_menuitem
 
     def make_numlockmenuitem(self):
         def numlock_toggled(*args):
@@ -402,8 +411,8 @@ class OSXMenuHelper(GTKTrayMenuBase):
                 self.numlock_menuitem.set_active(keyboard.num_lock_state)
             else:
                 log("set_numlock_menuitem(%s) no keyboard!", args)
-                self.swapkeys_menuitem.set_sensitive(False)
-        self.client.after_handshake(set_numlock_menuitem)
+                self.numlock_menuitem.set_sensitive(False)
+        self.after_handshake(set_numlock_menuitem)
         return self.numlock_menuitem
 
     def update_numlock(self, on):
@@ -425,10 +434,10 @@ class OSXMenuHelper(GTKTrayMenuBase):
             pixbuf = get_icon(icon_name)
             log("get_image(%s, %s) pixbuf=%s", icon_name, size, pixbuf)
             if not pixbuf:
-                return  None
+                return None
             if size:
                 return scaled_image(pixbuf, size)
-            return gtk.image_new_from_pixbuf(pixbuf)
+            return Gtk.Image.new_from_pixbuf(pixbuf)
         except Exception:
             log.error("get_image(%s, %s)", icon_name, size, exc_info=True)
-            return  None
+            return None
