@@ -1,6 +1,6 @@
 #!/usr/bin/env python3
 # This file is part of Xpra.
-# Copyright (c) 2016-2017 Antoine Martin <antoine@xpra.org>
+# Copyright (c) 2016-2023 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -10,6 +10,7 @@ import time
 import errno
 from ctypes import addressof, byref, c_ulong, c_char_p, c_char, c_void_p, cast, string_at
 from ctypes.wintypes import DWORD
+from typing import Dict, Any, Union
 
 from xpra.os_util import strtobytes, memoryview_to_bytes
 from xpra.net.bytestreams import Connection
@@ -69,7 +70,7 @@ class NamedPipeConnection(Connection):
         self.write_overlapped.InternalHigh = None
         self.write_overlapped.union.Pointer = None
 
-    def can_retry(self, e):
+    def can_retry(self, e) -> Union[bool,str]:
         code = e.args[0]
         if code==errno.WSAEWOULDBLOCK:      #@UndefinedVariable pylint: disable=no-member
             return "WSAEWOULDBLOCK"
@@ -108,13 +109,13 @@ class NamedPipeConnection(Connection):
             r = WaitForSingleObject(self.read_event, INFINITE)
             log("WaitForSingleObject(..)=%s, len=%s", WAIT_STR.get(r, r), read.value)
             if r and self.pipe_handle:
-                raise Exception("failed to read from named pipe handle %s" % self.pipe_handle)
+                raise RuntimeError("failed to read from named pipe handle %s" % self.pipe_handle)
         if self.pipe_handle:
             if not GetOverlappedResult(self.pipe_handle, byref(self.read_overlapped), byref(read), False):
                 e = GetLastError()
                 if e in CONNECTION_CLOSED_ERRORS:
                     raise ConnectionClosedException(CONNECTION_CLOSED_ERRORS[e])
-                raise Exception("overlapped read failed: %s" % IO_ERROR_STR.get(e, e))
+                raise RuntimeError("overlapped read failed: %s" % IO_ERROR_STR.get(e, e))
         if read.value==0:
             data = None
         else:
@@ -122,7 +123,7 @@ class NamedPipeConnection(Connection):
         log("pipe_read: %i bytes", len(data or ""))          #, binascii.hexlify(s))
         return data
 
-    def write(self, buf):
+    def write(self, buf, packet_type=None):
         return self._write(self._pipe_write, buf)
 
     def _pipe_write(self, buf):
@@ -144,11 +145,11 @@ class NamedPipeConnection(Connection):
                 #closed already!
                 return written.value
             if r:
-                raise Exception("failed to write buffer to named pipe handle %s" % self.pipe_handle)
+                raise RuntimeError("failed to write buffer to named pipe handle %s" % self.pipe_handle)
         if self.pipe_handle:
             if not GetOverlappedResult(self.pipe_handle, byref(self.write_overlapped), byref(written), False):
                 e = GetLastError()
-                raise Exception("overlapped write failed: %s" % IO_ERROR_STR.get(e, e))
+                raise RuntimeError("overlapped write failed: %s" % IO_ERROR_STR.get(e, e))
             log("pipe_write: %i bytes written", written.value)
             if self.pipe_handle:
                 FlushFileBuffers(self.pipe_handle)
@@ -185,8 +186,8 @@ class NamedPipeConnection(Connection):
     def __repr__(self):
         return self.target
 
-    def get_info(self) -> dict:
-        d = Connection.get_info(self)
+    def get_info(self) -> Dict[str,Any]:
+        d = super().get_info()
         d["type"] = "named-pipe"
         d["closed"] = self.pipe_handle is None
         return d
@@ -197,7 +198,7 @@ def connect_to_namedpipe(pipe_name, timeout=10):
     start = time.time()
     while True:
         if time.time()-start>=timeout:
-            raise Exception("timeout waiting for named pipe '%s'" % pipe_name)
+            raise RuntimeError("timeout waiting for named pipe '%s'" % pipe_name)
         pipe_handle = CreateFileA(strtobytes(pipe_name), GENERIC_READ | GENERIC_WRITE,
                                   0, None, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, 0)
         log("CreateFileA(%s)=%#x", pipe_name, pipe_handle)
@@ -207,9 +208,9 @@ def connect_to_namedpipe(pipe_name, timeout=10):
         log("CreateFileA(..) error=%s", err)
         if err==ERROR_PIPE_BUSY:
             if WaitNamedPipeA(pipe_name, timeout*10000)==0:
-                raise Exception("timeout waiting for named pipe '%s'" % pipe_name)
+                raise RuntimeError("timeout waiting for named pipe '%s'" % pipe_name)
         else:
-            raise Exception("cannot open named pipe '%s': %s" % (pipe_name, FormatMessageSystem(err)))
+            raise RuntimeError("cannot open named pipe '%s': %s" % (pipe_name, FormatMessageSystem(err)))
     #we have a valid handle!
     dwMode = c_ulong(PIPE_READMODE_BYTE)
     r = SetNamedPipeHandleState(pipe_handle, byref(dwMode), None, None)
