@@ -1,12 +1,13 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
-# Copyright (C) 2012-2019 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2012-2023 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 # Simple statistical functions
 
 import math
+from typing import List, Optional
 
 def to_std_unit(v, unit=1000):
     if v>=unit**3:
@@ -17,7 +18,7 @@ def to_std_unit(v, unit=1000):
         return "K", v//unit
     return "", v
 
-def std_unit(v, unit=1000):
+def std_unit(v, unit=1000) -> str:
     unit, value = to_std_unit(v, unit)
     return "%s%s" % (int(value), unit)
 
@@ -48,7 +49,7 @@ def values_to_scaled_values(data, scale_unit=10, min_scaled_value=10, num_values
     if not data:
         return 0, data
     max_v = max(data)
-    #pad with None values so we have at least num_values:
+    # pad with None values:
     if len(data)<num_values:
         if isinstance(data, tuple):
             data = list(data)
@@ -60,7 +61,7 @@ def values_to_scaled_values(data, scale_unit=10, min_scaled_value=10, num_values
         scale *= scale_unit
     if scale==1:
         return scale, data
-    sdata = []
+    sdata : List[Optional[float]] = []
     for x in data:
         if x is None:
             sdata.append(None)
@@ -101,10 +102,11 @@ def find_invpow(x, n):
     """Finds the integer component of the n'th root of x,
     an integer such that y ** n <= x < (y + 1) ** n.
     """
-    high = 1
+    high = 1.0
     while high ** n < x:
         high *= 2
     low = high/2
+    mid = low
     while low < high:
         mid = (low + high) // 2
         if low < mid and mid**n < x:
@@ -150,7 +152,7 @@ def get_list_stats(in_values, show_percentile=(5, 8, 9), show_dev=False):
             #geometric mean
             try:
                 v = int(math.pow(p, 1.0/counter))
-            except OverflowError:
+            except (ValueError, OverflowError):
                 v = find_invpow(p, counter)
             lstats["gm"] = v
         if h!=0:
