@@ -1,18 +1,24 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
 # Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
-# Copyright (C) 2010-2021 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2023 Antoine Martin <antoine@xpra.org>
 # Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 import os
+from typing import Dict, Any, List, Tuple
 
 from xpra.os_util import bytestostr, strtobytes, hexstr
-from xpra.util import nonl, typedict, envbool, iround
+from xpra.util import typedict, envbool
 from xpra.gtk_common.error import xswallow, xsync, xlog
+from xpra.scripts.config import parse_bool
+from xpra.server import EXITING_CODE
+from xpra.common import SYNC_ICC
 from xpra.x11.x11_server_core import X11ServerCore, XTestPointerDevice
-from xpra.x11.bindings.keyboard_bindings import X11KeyboardBindings #@UnresolvedImport
+from xpra.x11.bindings.keyboard import X11KeyboardBindings #@UnresolvedImport
+from xpra.x11.gtk_x11.prop import prop_set, prop_del
+from xpra.x11.xsettings_prop import XSettingsType, BLACKLISTED_XSETTINGS
 from xpra.log import Logger
 
 log = Logger("x11", "server")
@@ -23,10 +29,16 @@ dbuslog = Logger("dbus")
 X11Keyboard = X11KeyboardBindings()
 
 SCALED_FONT_ANTIALIAS = envbool("XPRA_SCALED_FONT_ANTIALIAS", False)
-SYNC_ICC = envbool("XPRA_SYNC_ICC", True)
 
 
-def _get_antialias_hintstyle(antialias):
+def root_prop_set(prop_name:str, prop_type, value) -> None:
+    # pylint: disable=import-outside-toplevel
+    prop_set(X11Keyboard.get_root_xid(), prop_name, prop_type, value)
+
+def root_prop_del(prop_name:str):
+    prop_del(X11Keyboard.get_root_xid(), prop_name)
+
+def _get_antialias_hintstyle(antialias:typedict) -> str:
     hintstyle = antialias.strget("hintstyle", "").lower()
     if hintstyle in ("hintnone", "hintslight", "hintmedium", "hintfull"):
         #X11 clients can give us what we need directly:
@@ -49,105 +61,120 @@ class X11ServerBase(X11ServerCore):
         (see XpraServer or DesktopServer for actual implementations)
     """
 
-    def __init__(self, clobber):
-        self.clobber = clobber
-        X11ServerCore.__init__(self)
-        self._default_xsettings = {}
-        self._settings = {}
-        self.double_click_time = 0
-        self.double_click_distance = 0
-        self.dpi = 0
-        self.default_dpi = 0
+    def __init__(self):
+        super().__init__()
+        self._default_xsettings : Tuple[int,List[Tuple]] = (0, [])
+        self._settings : Dict[str,Any] = {}
+        self.double_click_time : int = 0
+        self.double_click_distance : int = 0
+        self.dpi : int = 0
+        self.default_dpi : int = 0
         self._xsettings_manager = None
-        self._xsettings_enabled = False
-        self.display_pid = 0
+        self._xsettings_enabled : bool = False
+        self.display_pid : int = 0
+        self.icc_profile : bytes = b""
+        self.input_devices = "xtest"
 
-    def do_init(self, opts):
-        X11ServerCore.do_init(self, opts)
-        self._xsettings_enabled = opts.xsettings
+    def do_init(self, opts) -> None:
+        super().do_init(opts)
+        #the server class sets the default value for 'xsettings_enabled'
+        #it is overridden in the seamless server (enabled by default),
+        #and we let the options have the final say here:
+        self._xsettings_enabled = parse_bool("xsettings", opts.xsettings, self._xsettings_enabled)
+        log("xsettings_enabled(%s)=%s", opts.xsettings, self._xsettings_enabled)
         if self._xsettings_enabled:
             from xpra.x11.xsettings import XSettingsHelper
             self._default_xsettings = XSettingsHelper().get_settings()
             log("_default_xsettings=%s", self._default_xsettings)
             self.init_all_server_settings()
 
+    # noinspection PyMethodMayBeStatic
+    def save_pid(self) -> None:
+        root_prop_set("XPRA_SERVER_PID", "u32", os.getpid())
 
-    def init_display_pid(self, pid):
-        if pid:
-            from xpra.scripts.server import _save_int
-            _save_int(b"_XPRA_SERVER_PID", pid)
-        elif self.clobber:
-            from xpra.scripts.server import _get_int
-            pid = _get_int(b"_XPRA_SERVER_PID")
-            if not pid:
-                log.info("xvfb pid not found")
-            else:
-                log.info("xvfb pid=%i", pid)
+    def clean_x11_properties(self) -> None:
+        super().clean_x11_properties()
+        self.do_clean_x11_properties("XPRA_SERVER_PID")
+
+
+    def init_display_pid(self, pid:int) -> None:
+        if not pid:
+            log.info("xvfb pid not found")
+        else:
+            log.info(f"xvfb pid={pid}")
         self.display_pid = pid
 
-    def kill_display(self):
+    def kill_display(self) -> None:
         if self.display_pid:
-            from xpra.server import EXITING_CODE
             if self._upgrading==EXITING_CODE:
                 log.info("exiting: not cleaning up Xvfb")
             elif self._upgrading:
                 log.info("upgrading: not cleaning up Xvfb")
             else:
-                from xpra.scripts.server import add_cleanup
                 from xpra.x11.vfb_util import kill_xvfb
-                def do_kill_display():
-                    kill_xvfb(self.display_pid)
-                add_cleanup(do_kill_display)
+                kill_xvfb(self.display_pid)
+                self.do_clean_session_files(
+                    "xvfb.pid",
+                    "xauthority",
+                    "Xorg.log",
+                    "Xorg.log.old",
+                    "xorg.conf.d/*"
+                    "xorg.conf.d"
+                    )
 
 
-    def do_cleanup(self):
-        X11ServerCore.do_cleanup(self)
+    def late_cleanup(self) -> None:
+        super().late_cleanup()
         self.kill_display()
 
 
-    def configure_best_screen_size(self):
-        root_w, root_h = X11ServerCore.configure_best_screen_size(self)
+    def configure_best_screen_size(self) -> Tuple[int,int]:
+        root_w, root_h = super().configure_best_screen_size()
         if self.touchpad_device:
             self.touchpad_device.root_w = root_w
             self.touchpad_device.root_h = root_h
         return root_w, root_h
 
 
-    def init_dbus(self, dbus_pid, dbus_env):
-        try:
-            from xpra.server.dbus import dbus_start
-            assert dbus_start
-        except ImportError as e:
-            log("init_dbus(%s, %s)", dbus_pid, dbus_env, exc_info=True)
-            log.warn("Warning: cannot initialize dbus")
-            log.warn(" %s", e)
-            return
-        from xpra.server.dbus.dbus_start import (
-            get_saved_dbus_pid, get_saved_dbus_env,
-            save_dbus_pid, save_dbus_env,
-            )
-        super(X11ServerBase, self).init_dbus(dbus_pid, dbus_env)
-        if self.clobber:
-            #get the saved pids and env
-            self.dbus_pid = get_saved_dbus_pid()
-            self.dbus_env = get_saved_dbus_env()
-            dbuslog("retrieved existing dbus attributes: %s, %s", self.dbus_pid, self.dbus_env)
-            if self.dbus_env:
-                os.environ.update(self.dbus_env)
-        else:
+    def init_dbus(self, dbus_pid:int, dbus_env:Dict[str,str]) -> None:
+        dbuslog("init_dbus(%s, %s)", dbus_pid, dbus_env)
+        if dbus_pid and dbus_env:
+            os.environ.update(dbus_env)
+            self.dbus_pid = dbus_pid
+            self.dbus_env = dbus_env
             #now we can save values on the display
             #(we cannot access gtk3 until dbus has started up)
-            if self.dbus_pid:
-                save_dbus_pid(self.dbus_pid)
-            if self.dbus_env:
-                save_dbus_env(self.dbus_env)
+            def _save_int(prop_name, pid):
+                root_prop_set(prop_name, "u32", pid)
+            def _save_str(prop_name, s):
+                root_prop_set(prop_name, "latin1", s)
+            #DBUS_SESSION_BUS_ADDRESS=unix:abstract=/tmp/dbus-B8CDeWmam9,guid=b77f682bd8b57a5cc02f870556cbe9e9
+            #DBUS_SESSION_BUS_PID=11406
+            #DBUS_SESSION_BUS_WINDOWID=50331649
+            for n,conv,save in (
+                    ("ADDRESS",     bytestostr,     _save_str),
+                    ("PID",         int,            _save_int),
+                    ("WINDOW_ID",   int,            _save_int)):
+                k = f"DBUS_SESSION_BUS_{n}"
+                v = dbus_env.get(k)
+                if v is None:
+                    continue
+                try:
+                    tv = conv(v)
+                    save(k, tv)
+                except Exception as e:
+                    log("save_dbus_env(%s)", dbus_env, exc_info=True)
+                    log.error(f"Error: failed to save dbus environment variable {k!r}")
+                    log.error(f" with value {v!r}")
+                    log.estr(e)
 
 
-    def last_client_exited(self):
+    def last_client_exited(self) -> None:
         self.reset_settings()
-        X11ServerCore.last_client_exited(self)
+        super().last_client_exited()
 
-    def init_virtual_devices(self, devices):
+    def init_virtual_devices(self, devices:Dict[str,Any]) -> None:
+        # pylint: disable=import-outside-toplevel
         #(this runs in the main thread - before the main loop starts)
         #for the time being, we only use the pointer if there is one:
         pointer = devices.get("pointer")
@@ -174,13 +201,13 @@ class X11ServerBase(X11ServerCore):
         except Exception as e:
             mouselog("cannot get pointer device class from %s: %s", self.pointer_device, e)
 
-    def verify_uinput_pointer_device(self):
+    def verify_uinput_pointer_device(self) -> None:
         xtest = XTestPointerDevice()
         ox, oy = 100, 100
         with xlog:
-            xtest.move_pointer(0, ox, oy)
+            xtest.move_pointer(ox, oy)
         nx, ny = 200, 200
-        self.pointer_device.move_pointer(0, nx, ny)
+        self.pointer_device.move_pointer(nx, ny)
         def verify_uinput_moved():
             pos = None  #@UnusedVariable
             with xswallow:
@@ -189,18 +216,34 @@ class X11ServerBase(X11ServerCore):
             if pos==(ox, oy):
                 mouselog.warn("Warning: %s failed verification", self.pointer_device)
                 mouselog.warn(" expected pointer at %s, now at %s", (nx, ny), pos)
-                mouselog.warn(" usign XTest fallback")
+                mouselog.warn(" using XTest fallback")
                 self.pointer_device = xtest
                 self.input_devices = "xtest"
         self.timeout_add(1000, verify_uinput_moved)
 
 
-    def dpi_changed(self):
+    def dpi_changed(self) -> None:
         #re-apply the same settings, which will apply the new dpi override to it:
         self.update_server_settings()
 
 
-    def set_icc_profile(self):
+    def get_info(self, proto=None, client_uuids=None) -> Dict[str,Any]:
+        info = super().get_info(proto=proto, client_uuids=client_uuids)
+        display_info = info.setdefault("display", {})
+        if self.display_pid:
+            display_info["pid"] = self.display_pid
+        display_info["icc"] = self.get_icc_info()
+        return info
+
+    def get_icc_info(self) -> Dict[str,Any]:
+        icc_info : Dict[str,Any] = {
+            "sync"  : SYNC_ICC,
+            }
+        if SYNC_ICC:
+            icc_info["profile"] = hexstr(self.icc_profile)
+        return icc_info
+
+    def set_icc_profile(self) -> None:
         if not SYNC_ICC:
             return
         ui_clients = [s for s in self._server_sources.values() if s.ui_client]
@@ -209,41 +252,32 @@ class X11ServerBase(X11ServerCore):
             self.reset_icc_profile()
             return
         icc = typedict(ui_clients[0].icc)
-        data = None
         for x in ("data", "icc-data", "icc-profile"):
             data = icc.strget(x)
             if data:
-                break
-        if not data:
-            screenlog("no icc data found in %s", icc)
-            self.reset_icc_profile()
-            return
-        screenlog("set_icc_profile() icc data for %s: %s (%i bytes)",
-                  ui_clients[0], hexstr(data or ""), len(data or ""))
-        from xpra.x11.gtk_x11.prop import prop_set
-        #each CARD32 contains just one 8-bit value - don't ask me why
-        def o(x):
-            try:
-                return ord(x)
-            except:
-                return x
-        prop_set(self.root_window, "_ICC_PROFILE", ["u32"], [o(x) for x in data])
-        prop_set(self.root_window, "_ICC_PROFILE_IN_X_VERSION", "u32", 0*100+4) #0.4 -> 0*100+4*1
+                screenlog("set_icc_profile() icc data for %s: %s (%i bytes)",
+                          ui_clients[0], hexstr(data or ""), len(data or ""))
+                self.icc_profile = data
+                root_prop_set("_ICC_PROFILE", ["u32"], [ord(x) for x in data])
+                root_prop_set("_ICC_PROFILE_IN_X_VERSION", "u32", 0 * 100 + 4)  # 0.4 -> 0*100+4*1
+                return
+        screenlog("no icc data found in %s", icc)
+        self.reset_icc_profile()
 
-    def reset_icc_profile(self):
+    def reset_icc_profile(self) -> None:
         screenlog("reset_icc_profile()")
-        from xpra.x11.gtk_x11.prop import prop_del
-        prop_del(self.root_window, "_ICC_PROFILE")
-        prop_del(self.root_window, "_ICC_PROFILE_IN_X_VERSION")
+        root_prop_del("_ICC_PROFILE")
+        root_prop_del("_ICC_PROFILE_IN_X_VERSION")
+        self.icc_profile = b""
 
 
-    def reset_settings(self):
+    def reset_settings(self) -> None:
         if not self._xsettings_enabled:
             return
         log("resetting xsettings to: %s", self._default_xsettings)
         self.set_xsettings(self._default_xsettings or (0, ()))
 
-    def set_xsettings(self, v):
+    def set_xsettings(self, v) -> None:
         if not self._xsettings_enabled:
             return
         log("set_xsettings(%s)", v)
@@ -253,29 +287,31 @@ class X11ServerBase(X11ServerCore):
                 self._xsettings_manager = XSettingsManager()
             self._xsettings_manager.set_settings(v)
 
-    def init_all_server_settings(self):
+    def init_all_server_settings(self) -> None:
         log("init_all_server_settings() dpi=%i, default_dpi=%i", self.dpi, self.default_dpi)
         #almost like update_all, except we use the default_dpi,
         #since this is called before the first client connects
         self.do_update_server_settings({
-            b"resource-manager"  : b"",
-            b"xsettings-blob"    : (0, [])
+            "resource-manager"  : b"",
+            "xsettings-blob"    : (0, [])
             }, reset = True, dpi = self.default_dpi, cursor_size=24)
 
-    def update_all_server_settings(self, reset=False):
+    def update_all_server_settings(self, reset=False) -> None:
         self.update_server_settings({
-            b"resource-manager"  : b"",
-            b"xsettings-blob"    : (0, []),
+            "resource-manager"  : b"",
+            "xsettings-blob"    : (0, []),
             }, reset=reset)
 
-    def update_server_settings(self, settings=None, reset=False):
+    def update_server_settings(self, settings=None, reset=False) -> None:
         self.do_update_server_settings(settings or self._settings, reset,
-                                self.dpi, self.double_click_time, self.double_click_distance, self.antialias, self.cursor_size)
+                                self.dpi, self.double_click_time, self.double_click_distance,
+                                self.antialias, self.cursor_size)
 
     def do_update_server_settings(self, settings, reset=False,
-                                  dpi=0, double_click_time=0, double_click_distance=(-1, -1), antialias={}, cursor_size=-1):
+                                  dpi=0, double_click_time=0, double_click_distance=(-1, -1),
+                                  antialias=None, cursor_size=-1) -> None:
         if not self._xsettings_enabled:
-            log("ignoring xsettings update: %s", settings)
+            log(f"ignoring xsettings update: {settings}")
             return
         if reset:
             #FIXME: preserve serial? (what happens when we change values which had the same serial?)
@@ -287,19 +323,21 @@ class X11ServerBase(X11ServerCore):
                     for _, prop_name, value, _ in self._default_xsettings[1]:
                         self._settings[prop_name] = value
                 except Exception as e:
-                    log("failed to parse %s", self._default_xsettings)
+                    log(f"failed to parse {self._default_xsettings}")
                     log.warn("Warning: failed to parse default XSettings:")
-                    log.warn(" %s", e)
+                    log.warn(f" {e}")
         old_settings = dict(self._settings)
-        log("server_settings: old=%s, updating with=%s", nonl(old_settings), nonl(settings))
-        log("overrides: dpi=%s, double click time=%s, double click distance=%s",
-            dpi, double_click_time, double_click_distance)
-        log("overrides: antialias=%s", antialias)
+        log("server_settings: old=%r, updating with=%r", old_settings, settings)
+        log("overrides: ")
+        log(f" dpi={dpi}")
+        log(f" double click time={double_click_time}, double click distance={double_click_distance}")
+        log(f" antialias={antialias}")
+        #older versions may send keys as "bytes":
+        settings = dict((bytestostr(k), v) for k,v in settings.items())
         self._settings.update(settings)
-        from xpra.x11.xsettings_prop import XSettingsTypeInteger, XSettingsTypeString, BLACKLISTED_XSETTINGS
         for k, v in settings.items():
             #cook the "resource-manager" value to add the DPI and/or antialias values:
-            if k==b"resource-manager" and (dpi>0 or antialias or cursor_size>0):
+            if k=="resource-manager" and (dpi>0 or antialias or cursor_size>0):
                 value = bytestostr(v)
                 #parse the resources into a dict:
                 values={}
@@ -309,10 +347,10 @@ class X11ServerBase(X11ServerCore):
                         continue
                     parts = option.split(":\t", 1)
                     if len(parts)!=2:
-                        log("skipped invalid option: '%s'", option)
+                        log(f"skipped invalid option: {option!r}")
                         continue
                     if parts[0] in BLACKLISTED_XSETTINGS:
-                        log("skipped blacklisted option: '%s'", option)
+                        log(f"skipped blacklisted option: {option!r}")
                         continue
                     values[parts[0]] = parts[1]
                 if cursor_size>0:
@@ -327,7 +365,7 @@ class X11ServerBase(X11ServerCore):
                     sss = tuple(self._server_sources.values())
                     if len(sss)==1:
                         #only honour sub-pixel hinting if a single client is connected
-                        #and only when it is not using any scaling (or overriden with SCALED_FONT_ANTIALIAS):
+                        #and only when it is not using any scaling (or overridden with SCALED_FONT_ANTIALIAS):
                         ss = sss[0]
                         ds_unscaled = getattr(ss, "desktop_size_unscaled", None)
                         ds_scaled = getattr(ss, "desktop_size", None)
@@ -338,18 +376,18 @@ class X11ServerBase(X11ServerCore):
                                    "Xft.hinting"    : ad.intget("hinting", -1),
                                    "Xft.rgba"       : subpixel_order,
                                    "Xft.hintstyle"  : _get_antialias_hintstyle(ad)})
-                log("server_settings: resource-manager values=%s", nonl(values))
+                log(f"server_settings: resource-manager values={values}")
                 #convert the dict back into a resource string:
                 value = ''
                 for vk, vv in values.items():
-                    value += "%s:\t%s\n" % (vk, vv)
+                    value += f"{vk}:\t{vv}\n"
                 #record the actual value used
-                self._settings[b"resource-manager"] = value
+                self._settings["resource-manager"] = value
                 v = value.encode("utf-8")
 
             #cook xsettings to add various settings:
             #(as those may not be present in xsettings on some platforms.. like win32 and osx)
-            if k==b"xsettings-blob" and \
+            if k=="xsettings-blob" and \
             (self.double_click_time>0 or self.double_click_distance!=(-1, -1) or antialias or dpi>0):
                 #start by removing blacklisted options:
                 def filter_blacklisted():
@@ -365,42 +403,43 @@ class X11ServerBase(X11ServerCore):
                 def set_xsettings_value(name, value_type, value):
                     #remove existing one, if any:
                     serial, values = v
-                    new_values = [(_t,_n,_v,_s) for (_t,_n,_v,_s) in values if _n!=name]
-                    new_values.append((value_type, name, value, 0))
+                    bn = name.encode("utf-8")
+                    new_values = [(_t,_n,_v,_s) for (_t,_n,_v,_s) in values if _n!=bn]
+                    new_values.append((value_type, bn, value, 0))
                     return serial, new_values
                 def set_xsettings_int(name, value):
                     if value<0: #not set, return v unchanged
                         return v
-                    return set_xsettings_value(name, XSettingsTypeInteger, value)
+                    return set_xsettings_value(name, XSettingsType.Integer, value)
                 if dpi>0:
-                    v = set_xsettings_int(b"Xft/DPI", dpi*1024)
+                    v = set_xsettings_int("Xft/DPI", dpi*1024)
                 if double_click_time>0:
-                    v = set_xsettings_int(b"Net/DoubleClickTime", self.double_click_time)
+                    v = set_xsettings_int("Net/DoubleClickTime", double_click_time)
                 if antialias:
                     ad = typedict(antialias)
-                    v = set_xsettings_int(b"Xft/Antialias",  ad.intget("enabled", -1))
-                    v = set_xsettings_int(b"Xft/Hinting",    ad.intget("hinting", -1))
-                    v = set_xsettings_value(b"Xft/RGBA",     XSettingsTypeString, ad.strget("orientation", "none").lower())
-                    v = set_xsettings_value(b"Xft/HintStyle", XSettingsTypeString, _get_antialias_hintstyle(ad))
+                    v = set_xsettings_int("Xft/Antialias",  ad.intget("enabled", -1))
+                    v = set_xsettings_int("Xft/Hinting",    ad.intget("hinting", -1))
+                    orientation = ad.strget("orientation", "none").lower()
+                    v = set_xsettings_value("Xft/RGBA",     XSettingsType.String, orientation)
+                    v = set_xsettings_value("Xft/HintStyle", XSettingsType.String, _get_antialias_hintstyle(ad))
                 if double_click_distance!=(-1, -1):
                     #some platforms give us a value for each axis,
                     #but X11 only has one, so take the average
                     try:
                         x,y = double_click_distance
                         if x>0 and y>0:
-                            d = iround((x+y)/2.0)
+                            d = round((x+y)/2)
                             d = max(1, min(128, d))     #sanitize it a bit
-                            v = set_xsettings_int(b"Net/DoubleClickDistance", d)
+                            v = set_xsettings_int("Net/DoubleClickDistance", d)
                     except Exception as e:
                         log.warn("error setting double click distance from %s: %s", double_click_distance, e)
 
             if k not in old_settings or v != old_settings[k]:
-                if k == b"xsettings-blob":
+                if k == "xsettings-blob":
                     self.set_xsettings(v)
-                elif k == b"resource-manager":
-                    from xpra.x11.gtk_x11.prop import prop_set
+                elif k == "resource-manager":
                     p = "RESOURCE_MANAGER"
-                    log("server_settings: setting %s to %s", nonl(p), nonl(v))
-                    prop_set(self.root_window, p, "latin1", strtobytes(v).decode("latin1"))
+                    log(f"server_settings: setting {p} to {v}")
+                    root_prop_set(p, "latin1", strtobytes(v).decode("latin1"))
                 else:
-                    log.warn("Warning: unexpected setting '%s'", bytestostr(k))
+                    log.warn(f"Warning: unexpected setting {bytestostr(k)}")
